import numpy as np
import cv2
from shapely.geometry import Point, Polygon
from typing import List, Optional, Tuple

from ..geometry.chain import Chain
from ..analysis.chains_bag import ChainsBag
from ..processing.chain_context import ChainContext
from ..analysis.chain_system_manager import ChainSystemManager
from ..analysis.chain_analysis_tools import similarity_conditions
from ..analysis.interpolation_nodes import (
    complete_chain_using_2_support_ring,
    connect_2_chain_via_inward_and_outward_ring,
    complete_chain_using_support_ring,
)
from ..analysis.connect_chains import (
    intersection_between_chains,
    get_inward_and_outward_visible_chains,
)
from ..geometry.chain import Node, EndPoints, ClockDirection, TypeChains
from ..geometry.geometry_utils import (
    copy_node,
    copy_chain,
    get_nodes_angles_from_list_nodes,
    get_closest_dots_to_angle_on_radial_direction_sorted_by_ascending_distance_to_center,
    euclidean_distance_between_nodes,
    get_chain_from_list_by_id,
    visualize_selected_ch_and_chains_over_image_,
    angular_distance_between_chains,
    get_nodes_from_chain_list,
)
from ..config import config


def build_boundary_poly(outward_ring: Polygon, inward_ring: Polygon):
    """
    Convert shapely poly to region poly

    Args:
        outward_ring: outward shapely polygon
        inward_ring: inward shapely polygon

    Returns:
        Polygon region or None
    """
    if outward_ring is None and inward_ring is None:
        return None

    if outward_ring is not None and inward_ring is not None:
        x, y = outward_ring.geometry.exterior.coords.xy
        pts_ext = [[j, i] for i, j in zip(y, x)]
        x, y = inward_ring.geometry.exterior.coords.xy
        pts_int = [[j, i] for i, j in zip(y, x)]
        poly = Polygon(pts_ext, [pts_int])
    else:
        if outward_ring is None:
            x, y = inward_ring.geometry.exterior.coords.xy
        else:
            x, y = outward_ring.geometry.exterior.coords.xy

        pts_ext = [[j, i] for i, j in zip(y, x)]
        poly = Polygon(pts_ext)

    return poly


def search_shapely_inward_chain(
    shapley_incomplete_chain: Chain, outward_ring: Polygon, inward_ring: Polygon
):
    """
    Search for shapely polygon inside region delimited by outward and inward rings

    Args:
        shapley_incomplete_chain: shapely polygon chains not closed. Not nr nodes
        outward_ring: shapely polygon chain closed. nr nodes
        inward_ring: shapely polygon chain closed. nr nodes

    Returns:
        List of shapely inward chains subset
    """
    poly = build_boundary_poly(outward_ring, inward_ring)
    if poly is None:
        return []

    contains = np.vectorize(lambda p: poly.contains(Point(p)), signature="(n)->()")
    shapely_inward_chains_subset = []
    for cadena in shapley_incomplete_chain:
        x, y = cadena.xy
        pts = [[i, j] for i, j in zip(y, x)]
        if len(pts) == 0:
            continue
        try:
            vector = contains(np.ndarray(pts))
        except Exception as e:
            continue
        if outward_ring is not None:
            if vector.sum() == vector.shape[0]:
                shapely_inward_chains_subset.append(cadena)
        else:
            if vector.sum() == 0:
                shapely_inward_chains_subset.append(cadena)

    return shapely_inward_chains_subset


def from_shapely_to_chain(
    uncompleted_shapely_chain: List[Chain],
    uncomplete_chain: List[Chain],
    shapely_inward_chains_subset: List[Chain],
) -> List[Chain]:
    """
    Convert shapely chains to inward chains subset

    Args:
        uncompleted_shapely_chain: list of uncompleted shapely chains
        uncomplete_chain: list of uncompleted chains
        shapely_inward_chains_subset: list of shapely inward chains subset

    Returns:
        List of inward chains subset
    """
    inward_chain_subset = [
        uncomplete_chain[uncompleted_shapely_chain.index(cad_shapely)]
        for cad_shapely in shapely_inward_chains_subset
    ]
    inward_chain_subset.sort(key=lambda x: x.size, reverse=True)

    return inward_chain_subset


def select_support_chain(
    outward_ring: Optional[Chain], inward_ring: Optional[Chain], endpoint: Node
) -> Optional[Chain]:
    """
    Select the closest ring to the ch_j endpoint. Over endpoint ray direction, the support chain with the smallest
    distance between nodes is selected

    Args:
        outward_ring: outward support chain
        inward_ring: inward support chain
        endpoint: ch_j endpoint

    Returns:
        Closest support chain
    """
    chains_in_radial_direction = []

    if outward_ring is not None:
        chains_in_radial_direction.append(outward_ring)

    if inward_ring is not None:
        chains_in_radial_direction.append(inward_ring)

    dot_list_in_radial_direction = get_closest_dots_to_angle_on_radial_direction_sorted_by_ascending_distance_to_center(
        chains_in_radial_direction, endpoint.angle
    )

    distance = [
        euclidean_distance_between_nodes(endpoint, completed_chain_node)
        for completed_chain_node in dot_list_in_radial_direction
    ]

    if len(distance) < 2:
        support_chain = outward_ring if outward_ring is not None else inward_ring
    else:
        support_chain = get_chain_from_list_by_id(
            chains_in_radial_direction,
            dot_list_in_radial_direction[np.argmin(distance)].chain_id,
        )

    return support_chain


def exist_angular_intersection_with_src_chain(
    chain: Chain, src_chain_angular_domain: List[int]
) -> bool:
    """
    Check if there is an angular intersection with the source chain

    Args:
        chain: chain to check
        src_chain_angular_domain: angular domain of the source chain

    Returns:
        Boolean indicating if there is an angular intersection
    """
    domain = [node.angle for node in chain.l_nodes]
    if len(np.intersect1d(domain, src_chain_angular_domain)) == 0:
        return False

    return True


def angular_domain_overlapping_higher_than_threshold(
    src_chain_angular_domain: List[int],
    inter_chain: Chain,
    overlapping_threshold: int = 45,
) -> bool:
    """
    Check if overlapping angular domain between two chains is higher than a threshold

    Args:
        src_chain_angular_domain: angular domain of the source chain
        inter_chain: another chain
        overlapping_threshold: overlapping threshold

    Returns:
        Boolean indicating if the overlapping angular domain is higher than the threshold
    """
    inter_domain = [node.angle for node in inter_chain.l_nodes]
    inter = np.intersect1d(inter_domain, src_chain_angular_domain)
    rays_length = len(inter)
    angle_length = rays_length * 360 / inter_chain.nr
    if (len(inter) >= len(src_chain_angular_domain)) or (
        angle_length > overlapping_threshold
    ):
        return True
    else:
        return False


def split_chain(
    chain: Chain, node: Node, id_new: int = 10000000
) -> Tuple[Optional[Chain], Optional[Chain]]:
    """
    Split a chain into two chains

    Args:
        chain: Parent chain to be split
        node: Node element where the chain will be split
        id_new: New id for the split chains

    Returns:
        Tuple of child chains
    """
    node_list = chain.sort_dots()
    idx_split = node_list.index(node)
    ch1_node_list = [copy_node(node) for node in node_list[:idx_split]]
    if idx_split < len(node_list) - 1:
        ch2_node_list = [copy_node(node) for node in node_list[idx_split + 1 :]]
    else:
        ch2_node_list = []

    if len(ch1_node_list) > 1:
        ch1_sub = Chain(
            id_new,
            chain.nr,
            center=chain.center,
            img_height=chain.img_height,
            img_width=chain.img_width,
        )
        for node_ch in ch1_node_list:
            node_chain_id = ch1_sub.id

        ch1_sub.add_nodes_list(ch1_node_list)
    else:
        ch1_sub = None

    if len(ch2_node_list) > 1:
        ch2_sub = Chain(
            id_new,
            chain.nr,
            center=chain.center,
            img_height=chain.img_height,
            img_width=chain.img_width,
        )
        for node_ch in ch2_node_list:
            node_chain_id = ch2_sub.id
        ch2_sub.add_nodes_list(ch2_node_list)
    else:
        ch2_sub = None

    return (ch1_sub, ch2_sub)


def select_no_intersection_chain_at_endpoint(
    ch1_sub: Optional[Chain],
    ch2_sub: Optional[Chain],
    src_chain: Chain,
    ray_direction: int,
    total_nodes: int = 10,
) -> Optional[Chain]:
    """
    Select the chain that does not intersect with the source chain at endpoint

    Args:
        ch1_sub: Child chain 1
        ch2_sub: Child chain 2
        src_chain: Source chain
        ray_direction: Ray direction source chain
        total_nodes: Total nodes

    Returns:
        Chain that does not intersect with the source chain at endpoint
    """
    if ch1_sub is None and ch2_sub is None:
        return None
    if ch1_sub is None and ch2_sub is not None:
        return ch2_sub
    if ch2_sub is None and ch1_sub is not None:
        return ch1_sub

    endpoint = EndPoints.A if ray_direction == src_chain.extA.angle else EndPoints.B
    direction = (
        ClockDirection.clockwise
        if endpoint == EndPoints.B
        else ClockDirection.anti_clockwise
    )
    nodes_neighbourhood = src_chain.sort_dots(direction=direction)[:total_nodes]
    src_nodes = get_nodes_angles_from_list_nodes(nodes_neighbourhood)

    domain1 = (
        get_nodes_angles_from_list_nodes(ch1_sub.l_nodes)
        if ch1_sub.size > 0
        else src_nodes
    )
    domain2 = (
        get_nodes_angles_from_list_nodes(ch2_sub.l_nodes)
        if ch2_sub.size > 0
        else src_nodes
    )
    if np.intersect1d(domain1, src_nodes).shape[0] == 0:
        return ch1_sub
    elif np.intersect1d(domain2, src_nodes).shape[0] == 0:
        return ch2_sub
    else:
        return None


def split_intersecting_chains(
    direction: int, l_filtered_chains: List[Chain], ch_j: Chain
) -> List[Chain]:
    """
    Split intersecting chains. Implements Algorithm 18 in the supplementary material.

    Args:
        direction: Endpoint direction for split chains
        l_filtered_chains: List of chains to be split
        ch_j: Source chain

    Returns:
        List of split chains
    """
    l_search_chains = []
    for inter_chain in l_filtered_chains:
        split_node = inter_chain.get_node_by_angle(direction)
        if split_node is None:
            # It is not possible to split the chain due to split_node is None. Continue to next chain
            continue

        sub_ch1, sub_ch2 = split_chain(inter_chain, split_node)

        # Found what ch_i intersect the longest one
        ch_k = select_no_intersection_chain_at_endpoint(
            sub_ch1, sub_ch2, ch_j, direction
        )
        if ch_k is None:
            # There is not chain that does not intersect with ch_j at endpoint. Continue to next chain
            continue

        # Longest ch_i intersect two times
        if intersection_between_chains(ch_k, ch_j):
            node_direction_2 = (
                ch_j.extB.angle
                if split_node.angle == ch_j.extA.angle
                else ch_j.extA.angle
            )
            split_node_2 = ch_k.get_node_by_angle(node_direction_2)
            if split_node_2 is None:
                # It is not possible to split the chain due to split_node_2 is None. Continue to next chain
                continue

            sub_ch1, sub_ch2 = split_chain(ch_k, split_node_2)

            ch_k = select_no_intersection_chain_at_endpoint(
                sub_ch1, sub_ch2, ch_j, node_direction_2
            )
            if ch_k is None:
                # There is not chain that does not intersect with ch_j at endpoint. Continue to next chain
                continue

        ch_k.change_id(inter_chain.id)
        ch_k.label_id = inter_chain.label_id

        l_search_chains.append(ch_k)

    return l_search_chains


def split_intersecting_chain_in_other_endpoint(
    endpoint: EndPoints,
    src_chain: Chain,
    within_chain_set: List[Chain],
    within_nodes: List[Node],
    chain_search_set: List[Chain],
) -> List[Chain]:
    """
    Split intersecting chain in other endpoint

    Args:
        endpoint: Endpoint
        src_chain: Source chain
        within_chain_set: Chains within the region
        within_nodes: Nodes within the region
        chain_search_set: Chains to be split

    Returns:
        Updated chain search set
    """
    node_other_endpoint = src_chain.extB if endpoint == EndPoints.A else src_chain.extA
    direction = node_other_endpoint.angle
    node_direction = [
        node
        for node in within_nodes
        if ((node.angle == direction) and not (node.chain_id == src_chain.id))
    ]
    direction_cad_id = set([node.chain_id for node in node_direction])
    intersect_chain_id = [
        cad.id for cad in within_chain_set if cad.id in direction_cad_id
    ]
    intersecting_chains_in_other_endpoint = [
        chain for chain in chain_search_set if chain.id in intersect_chain_id
    ]
    chain_search_set = [
        chain
        for chain in chain_search_set
        if chain not in intersecting_chains_in_other_endpoint
    ]
    chain_search_set_in_other_endpoint = split_intersecting_chains(
        direction, intersecting_chains_in_other_endpoint, src_chain
    )
    chain_search_set += chain_search_set_in_other_endpoint

    return chain_search_set


def filter_no_intersected_chain_far(
    no_intersecting_chains: List[Chain],
    src_chain: Chain,
    endpoint: EndPoints,
    neighbourhood_size: int = 45,
) -> List[Chain]:
    """
    Filter the chains that are not intersected with the ch_j and are far from the endpoint

    Args:
        no_intersecting_chains: list of no intersecting chain with src chain
        src_chain: source chain
        endpoint: endpoint of source chain
        neighbourhood_size: angular neighbourhood size in degrees

    Returns:
        list of chains that are not intersected with the ch_j and are not far from the endpoint
    """
    closest_chains_set = []
    for chain in no_intersecting_chains:
        distance = angular_distance_between_chains(src_chain, chain, endpoint)
        if distance < neighbourhood_size:
            closest_chains_set.append((distance, chain))

    # sort by proximity to endpoint and return
    closest_chains_set.sort(key=lambda x: x[0])
    no_intersecting_chain_set = [chain for distance, chain in closest_chains_set]

    return no_intersecting_chain_set


def add_chains_that_intersect_in_other_endpoint(
    within_chain_set: List[Chain],
    no_intersections_chain: List[Chain],
    search_chain_set: List[Chain],
    src_chain: Chain,
    neighbourhood_size: int,
    endpoint: EndPoints,
) -> int:
    """
    Add chains that intersect in other endpoint

    Args:
        within_chain_set: chains in region
        no_intersections_chain: chains that do not intersect with ch_j.  ch_i that can be connected by this
        endpoint is added
        search_chain_set: candidate chains to be connected
        src_chain: source chan
        neighbourhood_size: neighbourhood size in degrees
        endpoint: source chain endpoint

    Returns:
        0
    """
    for in_chain in within_chain_set:
        if in_chain in no_intersections_chain + search_chain_set:
            continue
        if (
            angular_distance_between_chains(src_chain, in_chain, endpoint)
            < neighbourhood_size
        ):
            endpoint_in_chain = (
                in_chain.extA if EndPoints.A == endpoint else in_chain.extB
            )
            exist_intersection_in_other_endpoint = (
                src_chain.get_node_by_angle(endpoint_in_chain.angle) is not None
            )
            if exist_intersection_in_other_endpoint:
                # Check that there no intersection between both src endpoints
                sorted_order = (
                    ClockDirection.clockwise
                    if endpoint == EndPoints.A
                    else ClockDirection.anti_clockwise
                )
                in_chain_neighbourhood_nodes = in_chain.sort_dots(sorted_order)[
                    :neighbourhood_size
                ]
                if (
                    src_chain.get_node_by_angle(in_chain_neighbourhood_nodes[0].angle)
                    is None
                ):
                    search_chain_set.append(in_chain)
    return 0


def get_chains_that_satisfy_similarity_conditions(
    state: Optional[ChainSystemManager],
    support_chain: Chain,
    src_chain: Chain,
    search_chain_set: List[Chain],
    endpoint: EndPoints,
) -> Tuple[List[float], List[float], List[Chain]]:
    """
    Get chains that satisfy similarity conditions

    Args:
        state: debugging variable
        support_chain: support chain
        src_chain: source chain
        search_chain_set: list of candidate chains
        endpoint: source chain endpoint

    Returns:
        list of chain that satisfy similarity conditions
    """
    candidate_chains = []
    radial_distance_candidate_chains = []
    candidate_chain_euclidean_distance = []
    candidate_chain_idx = 0
    while True:
        if len(search_chain_set) <= candidate_chain_idx:
            break

        candidate_chain = search_chain_set[candidate_chain_idx]
        candidate_chain_idx += 1

        check_pass, distribution_distance = similarity_conditions(
            state=state,
            th_radial_tolerance=0.2,
            th_distribution_size=3,
            th_regular_derivative=2,
            derivative_from_center=False,
            ch_i=support_chain,
            ch_j=src_chain,
            candidate_chain=candidate_chain,
            endpoint=endpoint,
            check_overlapping=False,
        )
        if check_pass:
            candidate_chains.append(candidate_chain)
            radial_distance_candidate_chains.append(distribution_distance)

            endpoint_src = src_chain.extA if endpoint == EndPoints.A else src_chain.extB
            endpoint_candidate_chain = (
                candidate_chain.extB
                if endpoint == EndPoints.A
                else candidate_chain.extA
            )
            endpoint_distance = euclidean_distance_between_nodes(
                endpoint_src, endpoint_candidate_chain
            )
            candidate_chain_euclidean_distance.append(endpoint_distance)

    return (
        candidate_chain_euclidean_distance,
        radial_distance_candidate_chains,
        candidate_chains,
    )


def select_closest_candidate_chain(
    l_candidate_chains: List[Chain],
    l_candidate_chain_euclidean_distance: List[float],
    l_radial_distance_candidate_chains: List[float],
    l_within_chains: List[Chain],
    aux_chain: List[Chain],
) -> Tuple[List[Chain], float]:
    """
    Select closest chain by euclidean distance to ch_j chain.

    Args:
        l_candidate_chains (List): List of chain candidates.
        l_candidate_chain_euclidean_distance (List[float]): Euclidean distances of candidate chains.
        l_radial_distance_candidate_chains (List[float]): Radial distances of candidate chains.
        l_within_chains (List): Full list of chains within the region.
        aux_chain (List[Chain]): Check if the candidate chain is the same as aux_chain.

    Returns:
        Tuple[Optional, float]: Closest candidate chain by euclidean distance and radial distance to ch_j chain.
    """
    candidate_chain = None
    diff = -1

    if len(l_candidate_chains) > 0:
        candidate_chain = l_candidate_chains[
            np.argmin(l_candidate_chain_euclidean_distance)
        ]
        diff = np.min(l_radial_distance_candidate_chains)

    if aux_chain is not None and candidate_chain == aux_chain:
        if candidate_chain in l_within_chains:
            l_within_chains.remove(aux_chain)
            candidate_chain = None
            diff = -1

    return candidate_chain, diff


def select_nodes_within_region_over_ray(
    src_chain: Chain, endpoint_node, within_node_list: List[Node]
) -> List[Node]:
    """
    Select nodes within a region over a ray.

    Args:
        src_chain: Source chain.
        endpoint_node: Endpoint node.
        within_node_list: List of nodes within the region.

    Returns:
        List: Nodes within the region over the ray.
    """
    return [
        node
        for node in within_node_list
        if ((node.angle == endpoint_node.angle) and not (node.chain_id == src_chain.id))
    ]


def extract_chains_ids_from_nodes(nodes_ray: List[Node]) -> set[int]:
    """
    Extract chain IDs from nodes.

    Args:
        nodes_ray: List of nodes.

    Returns:
        set: Set of chain IDs.
    """
    return set([node.chain_id for node in nodes_ray])


def get_chains_from_ids(within_chains_set, chain_id_ray) -> List:
    """
    Get chains from IDs.

    Args:
        within_chains_set: Set of chains within the region.
        chain_id_ray: Set of chain IDs.

    Returns:
        List: List of chains corresponding to the IDs.
    """
    return [chain for chain in within_chains_set if chain.id in chain_id_ray]


def get_chains_that_no_intersect_src_chain(
    src_chain, src_chain_angle_domain, within_chains_set, endpoint_chain_intersections
) -> List:
    """
    Get chains that do not intersect with the source chain.

    Args:
        src_chain: Source chain.
        src_chain_angle_domain: Angle domain of the source chain.
        within_chains_set: Set of chains within the region.
        endpoint_chain_intersections: Chains that intersect with the endpoint.

    Returns:
        List: Chains that do not intersect with the source chain.
    """
    return [
        cad
        for cad in within_chains_set
        if cad not in endpoint_chain_intersections
        and cad != src_chain
        and not exist_angular_intersection_with_src_chain(cad, src_chain_angle_domain)
    ]


def remove_chains_with_higher_overlapping_threshold(
    src_chain_angle_domain, endpoint_chain_intersections, neighbourhood_size
) -> List:
    """
    Remove chains with higher overlapping threshold.

    Args:
        src_chain_angle_domain: Angle domain of the source chain.
        endpoint_chain_intersections: Chains that intersect with the endpoint.
        neighbourhood_size: Angular neighbourhood size in degrees.

    Returns:
        List: Chains with overlapping threshold lower than the specified size.
    """
    return [
        inter_chain
        for inter_chain in endpoint_chain_intersections
        if not angular_domain_overlapping_higher_than_threshold(
            src_chain_angle_domain,
            inter_chain,
            overlapping_threshold=neighbourhood_size,
        )
    ]


def remove_none_elements_from_list(list) -> List:
    """
    Remove None elements from a list.

    Args:
        list: Input list.

    Returns:
        List: List without None elements.
    """
    return [element for element in list if element is not None]


def split_and_connect_neighbouring_chains(
    l_within_nodes: List[Node],
    l_within_chains,
    ch_j: Chain,
    endpoint: int,
    outward_ring,
    inward_ring,
    neighbourhood_size,
    debug_params,
    save_path,
    aux_chain=None,
):
    """
    Logic for split and connect chains within region. Implements Algorithm 17 in the supplementary material.

    Args:
        l_within_nodes: nodes within region
        l_within_chains: chains within region
        ch_j: source chain. The one that is being to connect if condition are met.
        endpoint: endpoint of chain ch_j to find candidate chains to connect.
        outward_ring: outward support chain ring
        inward_ring: inward support chain ring
        neighbourhood_size: angular neighbourhood size in degrees to search for candidate chains
        debug_params: debug param
        save_path: debug param. Path to save debug images
        aux_chain: chain candidate to be connected by other endpoint. It is used to check that
        it is not connected by this endpoint

    Returns:
        candidate chain to connect, radial distance to ch_j and support chain.
    """
    img, iteration, debug = debug_params
    # Get angle domain for source ch_j
    ch_j_angle_domain = get_nodes_angles_from_list_nodes(ch_j.l_nodes)

    # Get endpoint node
    ch_j_node = ch_j.extA if endpoint == EndPoints.A else ch_j.extB

    # Select ch_j  support chain over endpoint
    ch_i = select_support_chain(outward_ring, inward_ring, ch_j_node)

    # Select within nodes over endpoint ray
    l_nodes_ray = select_nodes_within_region_over_ray(ch_j, ch_j_node, l_within_nodes)

    # Select within chains id over endpoint ray
    l_chain_id_ray = extract_chains_ids_from_nodes(l_nodes_ray)

    # Select within chains over endpoint ray by chain id
    l_endpoint_chains = get_chains_from_ids(l_within_chains, l_chain_id_ray)

    # filter chains that intersect with an overlapping threshold higher than 45 degrees. If overlapping threshold is
    # so big, it is not a good candidate to connect
    l_filtered_chains = remove_chains_with_higher_overlapping_threshold(
        ch_j_angle_domain, l_endpoint_chains, neighbourhood_size
    )

    if debug:
        boundary_ring_list = remove_none_elements_from_list([outward_ring, inward_ring])
        visualize_selected_ch_and_chains_over_image_(
            [ch_i, ch_j] + l_filtered_chains + boundary_ring_list,
            l_within_chains,
            img,
            f"{save_path}/{iteration[0]}_split_chains_{ch_j.label_id}_2_1_{endpoint}_{ch_i.label_id}.png",
        )
        iteration[0] += 1

    # Split intersection chains by endpoint. Algorithm 18 in the supplementary material
    l_candidates = split_intersecting_chains(ch_j_node.angle, l_filtered_chains, ch_j)
    if debug:
        visualize_selected_ch_and_chains_over_image_(
            [ch_i, ch_j] + l_candidates + boundary_ring_list,
            l_within_chains,
            img,
            f"{save_path}/{iteration[0]}_split_chains_{ch_j.label_id}_2_2_{endpoint}.png",
        )
        iteration[0] += 1

    # Select chains that do not intersect to ch_j
    l_no_intersection_j = get_chains_that_no_intersect_src_chain(
        ch_j, ch_j_angle_domain, l_within_chains, l_endpoint_chains
    )
    if aux_chain is not None:
        # If aux_chain is candidate to connect by the other endpoint, add it to the list of chains that do not intersect
        l_no_intersection_j += [aux_chain]

    # Add ch_i that intersect in other endpoint
    add_chains_that_intersect_in_other_endpoint(
        l_within_chains,
        l_no_intersection_j,
        l_candidates,
        ch_j,
        neighbourhood_size,
        endpoint,
    )
    if debug:
        visualize_selected_ch_and_chains_over_image_(
            [ch_i, ch_j] + l_candidates + boundary_ring_list,
            l_within_chains,
            img,
            f"{save_path}/{iteration[0]}_split_chains_{ch_j.label_id}_2_3_{endpoint}.png",
        )
        iteration[0] += 1

    # Split intersection chains by other endpoint
    l_candidates = split_intersecting_chain_in_other_endpoint(
        endpoint, ch_j, l_within_chains, l_within_nodes, l_candidates
    )
    if debug:
        visualize_selected_ch_and_chains_over_image_(
            [ch_i, ch_j] + l_candidates + boundary_ring_list,
            l_within_chains,
            img,
            f"{save_path}/{iteration[0]}_split_chains_{ch_j.label_id}_2_4_{endpoint}.png",
        )
        iteration[0] += 1

    # Filter no intersected chains that are far from endpoint
    l_candidates += filter_no_intersected_chain_far(
        l_no_intersection_j, ch_j, endpoint, neighbourhood_size
    )
    if debug:
        visualize_selected_ch_and_chains_over_image_(
            [ch_i, ch_j] + l_candidates + boundary_ring_list,
            l_within_chains,
            img,
            f"{save_path}/{iteration[0]}_split_chains_{ch_j.label_id}_2_5_{endpoint}.png",
        )
        iteration[0] += 1
        counter_init = iteration[0]
        state = ChainSystemManager(
            [ch_j],
            [ch_j.l_nodes],
            np.zeros((2, 2)),
            ch_j.center,
            debug=debug,
            counter=iteration[0],
            save=f"{save_path}",
            img=img,
        )

    else:
        state = None

    # Get chains that satisfy similarity conditions
    l_ch_k_euclidean_distance, l_ch_k_radial_distance, l_ch_k = (
        get_chains_that_satisfy_similarity_conditions(
            state, ch_i, ch_j, l_candidates, endpoint
        )
    )

    # Select ch_k candidate ch_i that satisfy similarity conditions
    ch_k, diff = select_closest_candidate_chain(
        l_ch_k,
        l_ch_k_euclidean_distance,
        l_ch_k_radial_distance,
        l_within_chains,
        aux_chain,
    )
    if debug:
        iteration[0] += state.counter - counter_init
        if ch_k is not None:
            visualize_selected_ch_and_chains_over_image_(
                [ch_i, ch_j] + [ch_k] + boundary_ring_list,
                l_within_chains,
                img,
                f"{save_path}/{iteration[0]}_split_chains_{ch_j.label_id}_2_6_2_{endpoint}.png",
            )
            iteration[0] += 1

    return ch_k, diff, ch_i


def debugging_postprocessing(
    debug, l_ch, img, l_within_chain_subset, filename, iteration
):
    if debug:
        visualize_selected_ch_and_chains_over_image_(
            l_ch, l_within_chain_subset, img, filename
        )
        iteration[0] += 1


def split_and_connect_chains(
    l_within_chains: List[Chain],
    inward_ring: Chain,
    outward_ring: Chain,
    l_ch_p: List[Chain],
    l_nodes_c: List[Node],
    neighbourhood_size=45,
    debug=False,
    img=None,
    save_path=None,
    iteration=None,
):
    """
    Split chains that intersect in other endpoint and connect them if connectivity goodness conditions are met.
    Implements Algorithm 16 in the supplementary material.

    Args:
        l_within_chains: uncompleted chains delimitated by inward_ring and outward_ring
        inward_ring: inward ring of the region.
        outward_ring: outward ring of the region.
        l_ch_p: full chain list
        l_nodes_c: full nodes list
        neighbourhood_size: total_nodes size to search for chains that intersect in other endpoint
        debug: Set to true if debugging is allowed
        img: debug parameter. Image matrix
        save_path: debug parameter. Path to save debugging images
        iteration: debug parameter. Iteration counter

    Returns:
        boolean value indicating if a chain was completed over region
    """
    # Line 1 Initialization step
    l_within_chains.sort(key=lambda x: x.size, reverse=True)
    connected = False
    completed_chain = False
    ch_j = None
    debug_params = img, iteration, debug

    # Line 2 Get inward nodes
    l_inward_nodes = get_nodes_from_chain_list(l_within_chains)

    # Line 3 Main loop to split chains that intersect over endpoints. Generator is defined to get next chain.
    generator = ChainsBag(l_within_chains)
    while True:
        # Line 4
        if not connected:
            if ch_j is not None and ch_j.is_closed(threshold=0.75):
                # Line 6
                complete_chain_using_2_support_ring(inward_ring, outward_ring, ch_j)
                completed_chain = True
                debugging_postprocessing(
                    debug,
                    [ch_j],
                    l_within_chains,
                    img,
                    f"{save_path}/{iteration[0]}_split_chains_{ch_j.label_id}.png",
                    iteration,
                )
                # Line 8
                ch_j = None

            else:
                # Line 10. Generator has pointers to l_within_chains list. It is used to get next chain
                ch_j = generator.get_next_chain()

        if ch_j is None:
            break
        debugging_postprocessing(
            debug,
            [ch_j, inward_ring, outward_ring],
            l_within_chains,
            img,
            f"{save_path}/{iteration[0]}_split_chains_{ch_j.label_id}_init.png",
            iteration,
        )

        # Line 13 Split chains in endpoint A and get candidate ch_i. Algorithm 17 in the supplementary material.
        endpoint = EndPoints.A
        ch_k_a, diff_a, ch_i_a = split_and_connect_neighbouring_chains(
            l_inward_nodes,
            l_within_chains,
            ch_j,
            endpoint,
            outward_ring,
            inward_ring,
            neighbourhood_size,
            debug_params,
            save_path=save_path,
        )
        # Line 14 Split chains in endpoint B and get candidate ch_i
        endpoint = EndPoints.B
        ch_k_b, diff_b, ch_i_b = split_and_connect_neighbouring_chains(
            l_inward_nodes,
            l_within_chains,
            ch_j,
            endpoint,
            outward_ring,
            inward_ring,
            neighbourhood_size,
            debug_params,
            save_path=save_path,
            aux_chain=ch_k_a,
        )
        # debug
        if debug:
            candidates_set = []
            if ch_k_b is not None:
                candidates_set.append(ch_k_b)
                candidates_set.append(ch_i_b)

            if ch_k_a is not None:
                candidates_set.append(ch_k_a)
                candidates_set.append(ch_i_a)
            debugging_postprocessing(
                debug,
                [ch_j] + candidates_set,
                l_within_chains,
                img,
                f"{save_path}/{iteration[0]}_split_chains_{ch_j.label_id}_candidate.png",
                iteration,
            )

        # Line 15
        connected, ch_i, endpoint = connect_radially_closest_chain(
            ch_j,
            ch_k_a,
            diff_a,
            ch_i_a,
            ch_k_b,
            diff_b,
            ch_i_b,
            l_ch_p,
            l_within_chains,
            l_nodes_c,
            inward_ring,
            outward_ring,
        )

        debugging_postprocessing(
            debug,
            [ch_i, ch_j],
            l_within_chains,
            img,
            f"{save_path}/{iteration[0]}_split_chains_{ch_j.label_id}_end.png",
            iteration,
        )

    # Line 16
    return completed_chain


def connect_2_chain_via_support_chain(
    outward_chain,
    inward_chain,
    src_chain,
    candidate_chain,
    nodes_list,
    endpoint,
    chain_list,
    inner_chain_list,
):
    """
    Connect 2 chains using outward and inward chain as support chains

    Args:
        outward_chain: outward support chain
        inward_chain: inward support chain
        src_chain: source chain
        candidate_chain: candidate chain
        nodes_list: full node list
        endpoint: source chain endpoint
        chain_list: full chain list
        inner_chain_list: chain list delimitated by inward_ring and outward_ring

    Returns:
        None. Chains are modified in place. Candidate ch_i is removed from chain_list and inner_chain_list and ch_j is modified
    """
    connect_2_chain_via_inward_and_outward_ring(
        outward_chain, inward_chain, src_chain, candidate_chain, nodes_list, endpoint
    )

    # Remove ch_i from ch_i lists. Candidate ch_i must be removed from inner_chain_list(region) and chain_list(global)
    inner_candidate_chain = get_chain_from_list_by_id(
        inner_chain_list, candidate_chain.id
    )
    if inner_candidate_chain is not None:
        cadena_ref_lista_original = inner_candidate_chain
        inner_chain_list.remove(cadena_ref_lista_original)
        chain_list.remove(cadena_ref_lista_original)

    global_candidate_chain = get_chain_from_list_by_id(chain_list, candidate_chain.id)
    if global_candidate_chain is not None:
        chain_list.remove(global_candidate_chain)

    return


def connect_radially_closest_chain(
    src_chain,
    candidate_chain_a,
    diff_a,
    support_chain_a,
    candidate_chain_b,
    diff_b,
    support_chain_b,
    ch_p_list,
    within_chains_subset,
    node_c_list,
    inward_ring,
    outward_ring,
):
    """
    Given 2 candidate chains, connect the one that is radially closer to the source chain

    Args:
        src_chain: source chain
        candidate_chain_a: candidate chain at endpoint A
        diff_a: difference between source chain and candidate chain at endpoint A
        support_chain_a: support chain at endpoint A
        candidate_chain_b: candidate chain at endpoint B
        diff_b: difference between source chain and candidate chain at endpoint B
        support_chain_b: support chain at endpoint B
        ch_p_list: full chain list over disk
        within_chains_subset: chains within the region of interest
        node_c_list: full node list over disk
        inward_ring: inward ring delimiting region of interest
        outward_ring: outward ring delimiting region of interest

    Returns:
        boolean value indicating if the chains were connected. If true, support chain and endpoint are returned
    """
    if (0 <= diff_a <= diff_b) or (diff_b < 0 and diff_a >= 0):
        candidate_chain = candidate_chain_a
        support_chain = support_chain_a
        endpoint = EndPoints.A

    elif (0 <= diff_b < diff_a) or (diff_a < 0 and diff_b >= 0):
        candidate_chain = candidate_chain_b
        support_chain = support_chain_b
        endpoint = EndPoints.B

    else:
        return False, support_chain_a, ""

    if candidate_chain.size + src_chain.size > candidate_chain.nr:
        return False, support_chain_a, ""

    connect_2_chain_via_support_chain(
        outward_ring,
        inward_ring,
        src_chain,
        candidate_chain,
        node_c_list,
        endpoint,
        ch_p_list,
        within_chains_subset,
    )

    return True, support_chain, endpoint


def postprocessing(l_ch_c, l_nodes_c, img_pre):
    """
    Posprocessing chain list. Conditions are relaxed in order to re-fine chain connections. Implements Algorithm 15
    in the supplementary material.

    Args:
        l_ch_c: chain list
        l_nodes_c: node list
        img_pre: input image

    Returns:
        connected chains list
    """
    l_ch_p = [copy_chain(chain) for chain in l_ch_c]
    chain_was_completed = False
    idx_start = None

    # debug parameter
    iteracion = [0]

    while True:
        ctx = ChainContext(
            l_ch_p, idx_start, save_path=str(config.output_dir), img=img_pre
        )

        while len(ctx.completed_chains) > 0:
            #  l_within_chains, inward_ring and outward_ring are attributes of ctx which are updated in next line
            ctx.update()
            if config.debug:
                ctx.drawing(iteracion[0])
                iteracion[0] += 1

            # First Postprocessing. Split all chains and connect them if it possible (Algorithm 16 in the supplementary material)
            chain_was_completed = split_and_connect_chains(
                ctx.l_within_chains,
                ctx.inward_ring,
                ctx.outward_ring,
                l_ch_p,
                l_nodes_c,
                neighbourhood_size=ctx.neighbourhood_size,
                debug=config.debug,
                img=img_pre,
                save_path=config.output_dir,
                iteration=iteracion,
            )
            # If ch_i was completed, restart iteration
            if chain_was_completed:
                idx_start = ctx.idx
                break

            # Second posproccessing
            connect_chains_if_there_is_enough_data(ctx, l_nodes_c, l_ch_p)

            if ctx.exit():
                break

        if not chain_was_completed:
            break

    # final Step, complete chains
    complete_chains_if_required(l_ch_p)

    return l_ch_p


def connect_chains_if_there_is_enough_data(ctx, l_nodes_c, l_ch_p):
    """
    Connect chains if there is enough data. This is the last step of the postprocessing

    Args:
        ctx: context object
        l_nodes_c: full node list in disk
        l_ch_p: full chain list in disk

    Returns:
        None. Chains are modified in place
    """
    there_is_chain = len(ctx.l_within_chains) == 1
    if there_is_chain:
        l_inward_chains = ctx.l_within_chains[0]
        postprocessing_unique_chain(
            l_inward_chains, ctx.inward_ring, ctx.outward_ring, l_nodes_c
        )
        return

    more_than_1_chain = len(ctx.l_within_chains) > 1
    if more_than_1_chain:
        postprocessing_more_than_one_chain_without_intersection(
            ctx.l_within_chains, ctx.inward_ring, ctx.outward_ring, l_nodes_c, l_ch_p
        )

    return 0


def complete_chains_if_required(ch_p):
    """
    Complete chains if full and size is less than nr

    Args:
        ch_p: chain list to complete

    Returns:
        None. Chains are modified in place
    """
    chain_list = [chain for chain in ch_p if chain.type not in [TypeChains.border]]
    for chain in chain_list:
        if chain.is_closed() and chain.size < chain.nr:
            inward_chain, outward_chain, _ = get_inward_and_outward_visible_chains(
                chain_list, chain, EndPoints.A.value
            )
            if inward_chain is not None and outward_chain is not None:
                complete_chain_using_2_support_ring(inward_chain, outward_chain, chain)

            elif inward_chain is not None or outward_chain is not None:
                support_chain = None
                complete_chain_using_support_ring(support_chain, chain)

    return 0


def postprocessing_unique_chain(
    within_chain,
    inward_ring_chain,
    outward_ring_chain,
    node_list,
    information_threshold=180,
):
    """
    Postprocessing for unique chain if chain size is greater than information threshold

    Args:
        within_chain: chain in region
        inward_ring_chain: inward ring chain
        outward_ring_chain: outward ring chain
        node_list: full node list in disk
        information_threshold: data threshold

    Returns:
        None. Chains are modified in place
    """
    within_chain_angular_size = within_chain.size * 360 / within_chain.nr
    if within_chain_angular_size > information_threshold:
        complete_chain_using_2_support_ring(
            inward_ring_chain, outward_ring_chain, within_chain
        )

    return


def build_no_intersecting_chain_set(chains_subset):
    """
    Build a set of chains that do not intersect with each other.

    Args:
        chains_subset: all chains within region

    Returns:
        subset of chains that do not intersect with each other
    """
    chains_subset.sort(key=lambda x: x.size)
    chains_subset = [chain for chain in chains_subset if not chain.is_closed()]
    no_intersecting_subset = []

    while len(chains_subset) > 0:
        longest_chain = chains_subset[-1]
        longest_chain_intersect_already_added_chain = (
            len(
                [
                    chain
                    for chain in no_intersecting_subset
                    if intersection_between_chains(chain, longest_chain)
                ]
            )
            > 0
        )

        chains_subset.remove(longest_chain)
        if longest_chain_intersect_already_added_chain:
            continue

        no_intersecting_subset.append(longest_chain)

    return no_intersecting_subset


def postprocessing_more_than_one_chain_without_intersection(
    chain_subset,
    outward_ring_chain,
    inward_ring_chain,
    node_list,
    chain_list,
    information_threshold=180,
):
    """
    Postprocessing for more than one chain without intersection. If we have more than one chain in region that
    not intersect each other. This chain subset also have to have an angular domain higher than information_threshold.
    Then we iterate over the chains and if satisfy similarity condition, we can connect them.

    Args:
        chain_subset: chains in region defined by outward and inward ring
        outward_ring_chain: outward ring chain
        inward_ring_chain: inward ring chain
        node_list: full node list in all the disk
        chain_list: full chain list in all the disk, not only the region
        information_threshold: threshold to connect chains in degrees

    Returns:
        connect chains if it possible
    """
    # get all the chains that not intersect each other
    no_intersecting_subset = build_no_intersecting_chain_set(chain_subset)
    angular_step = 360 / outward_ring_chain.nr
    enough_information = (
        np.sum([cad.size for cad in no_intersecting_subset]) * angular_step
        > information_threshold
    )
    if not enough_information:
        return 0

    no_intersecting_subset.sort(key=lambda x: x.extA.angle)

    # Fist chain. All the nodes of chain that satisfy similarity condition will be added to this chain
    src_chain = no_intersecting_subset.pop(0)
    endpoint_node = src_chain.extB
    endpoint = EndPoints.B
    # Select radially closer chain to ch_j endpoint
    support_chain = select_support_chain(
        outward_ring_chain, inward_ring_chain, endpoint_node
    )

    # Iterate over the rest of chains
    while len(no_intersecting_subset) > 0:
        next_chain = no_intersecting_subset[0]
        check_pass, distribution_distance = similarity_conditions(
            None,
            0.2,
            3,
            2,
            False,
            support_chain,
            src_chain,
            next_chain,
            endpoint,
            check_overlapping=True,
            chain_list=chain_subset,
        )

        if check_pass:
            # connect candidate_chain to ch_j
            connect_2_chain_via_support_chain(
                outward_ring_chain,
                inward_ring_chain,
                src_chain,
                next_chain,
                node_list,
                endpoint,
                chain_list,
                no_intersecting_subset,
            )
        else:
            no_intersecting_subset.remove(next_chain)

    complete_chain_using_2_support_ring(
        inward_ring_chain, outward_ring_chain, src_chain
    )

    return 0



from typing import Any, List, Optional, Tuple
import cv2
import matplotlib.pyplot as plt
import numpy as np

from ..geometry.virtual_band_generator import VirtualBandGenerator
from ..analysis.chain_neighbourhood import ChainNeighbourhood
from ..geometry.node import Node
from ..geometry.chain import Chain, EndPoints
from ..geometry.geometry_utils import (
    get_node_from_list_by_angle,
    visualize_selected_ch_and_chains_over_image_,
)
from ..analysis.interpolation_nodes import (
    domain_interpolation,
)


def draw_segment_between_nodes(
    pto1: Node,
    pto2: Node,
    img: np.ndarray,
    color: Tuple[int, int, int] = (0, 255, 0),
    thickness: int = 2,
) -> np.ndarray:
    """
    Draws a segment between two nodes on an image.

    Args:
        pto1 (Node): First point.
        pto2 (Node): Second point.
        img (np.ndarray): Image to draw on.
        color (Tuple[int, int, int]): Color of the line.
        thickness (int): Thickness of the line.

    Returns:
        np.ndarray: Image with the drawn segment.
    """
    pts = np.ndarray([[pto1.y, pto1.x], [pto2.y, pto2.x]], dtype=int)
    isClosed = False
    img = cv2.polylines(img, [pts], isClosed, color, thickness)

    return img


def vector_derivative(f: np.ndarray, nr: int) -> np.ndarray:
    """
    Computes the central derivative of a vector.

    Args:
        f (np.ndarray): Input vector.
        nr (int): Number of rays (unused).

    Returns:
        np.ndarray: Derivative of the vector.
    """
    return np.gradient(f)


def regularity_of_the_derivative_condition(
    state: Any,
    nr: int,
    ch_jk_nodes: List[Node],
    ch_j_nodes: List[Node],
    ch_k_nodes: List[Node],
    endpoint_j: EndPoints,
    th_deriv: float = 1.0,
) -> bool:
    """
    Computes the regularity of the derivative condition.

    Args:
        state (Any): State object for debugging purposes.
        nr (int): Number of rays.
        ch_jk_nodes (List[Node]): Nodes of the combined chains.
        ch_j_nodes (List[Node]): Nodes of the source chain.
        ch_k_nodes (List[Node]): Nodes of the destination chain.
        endpoint_j (EndPoints): Endpoint of the source chain.
        th_deriv (float): Derivative threshold.

    Returns:
        bool: True if the derivative is regular, False otherwise.
    """
    ch_jk_radials = [node.radial_distance for node in ch_jk_nodes]
    nodes_radial_distance_src_chain = [node.radial_distance for node in ch_j_nodes]
    nodes_radial_distance_dst_chain = [node.radial_distance for node in ch_k_nodes]

    abs_der_1 = np.abs(vector_derivative(nodes_radial_distance_src_chain, nr))
    abs_der_2 = np.abs(vector_derivative(nodes_radial_distance_dst_chain, nr))
    abs_der_3 = np.abs(vector_derivative(ch_jk_radials, nr))
    maximum_derivative_chains = np.maximum(abs_der_1.max(), abs_der_2.max())

    max_derivative_end = np.max(abs_der_3)
    RegularDeriv = max_derivative_end <= maximum_derivative_chains * th_deriv

    if state is not None and getattr(state, "debug", False):
        f, (ax2, ax1) = plt.subplots(2, 1)
        ax1.plot(abs_der_3)
        if endpoint_j == EndPoints.A:
            ax1.plot(np.arange(0, len(abs_der_2)), abs_der_2[::-1])
            ax1.plot(
                np.arange(
                    len(ch_jk_radials) - len(nodes_radial_distance_src_chain),
                    len(ch_jk_radials),
                ),
                abs_der_1,
            )
        else:
            ax1.plot(np.arange(0, len(abs_der_1)), abs_der_1[::-1])
            ax1.plot(
                np.arange(
                    len(ch_jk_radials) - len(nodes_radial_distance_dst_chain),
                    len(ch_jk_radials),
                ),
                abs_der_2,
            )

        ax1.hlines(
            y=max_derivative_end,
            xmin=0,
            xmax=np.maximum(
                len(nodes_radial_distance_src_chain),
                len(nodes_radial_distance_dst_chain),
            ),
            label="Salto",
        )
        ax1.hlines(
            y=th_deriv * maximum_derivative_chains,
            xmin=0,
            xmax=np.maximum(
                len(nodes_radial_distance_dst_chain),
                len(nodes_radial_distance_src_chain),
            ),
            colors="r",
            label="umbral",
        )
        ax1.legend()

        ax2.plot(ch_jk_radials)
        if endpoint_j == EndPoints.A:
            ax2.plot(
                np.arange(0, len(abs_der_2)), nodes_radial_distance_dst_chain[::-1], "r"
            )
            ax2.plot(
                np.arange(
                    len(ch_jk_radials) - len(nodes_radial_distance_src_chain),
                    len(ch_jk_radials),
                ),
                nodes_radial_distance_src_chain,
            )
        else:
            ax2.plot(
                np.arange(0, len(abs_der_1)), nodes_radial_distance_src_chain[::-1], "r"
            )
            ax2.plot(
                np.arange(
                    len(ch_jk_radials) - len(nodes_radial_distance_dst_chain),
                    len(ch_jk_radials),
                ),
                nodes_radial_distance_dst_chain,
            )

        plt.title(f"{RegularDeriv}")
        plt.savefig(f"{str(state.path)}/{state.counter}_derivada_{RegularDeriv}.png")
        plt.close()
        state.counter += 1

    return RegularDeriv


def generate_virtual_nodes_without_support_chain(
    ch_1: Chain, ch_2: Chain, endpoint: EndPoints
) -> List[Node]:
    """
    Generates virtual nodes between two chains without a support chain.

    Args:
        ch_1 (Chain): First chain.
        ch_2 (Chain): Second chain.
        endpoint (EndPoints): Endpoint to connect.

    Returns:
        List[Node]: List of virtual nodes.
    """
    ch1_border = ch_1.extA if endpoint == EndPoints.A else ch_1.extB
    ch2_border = ch_2.extB if endpoint == EndPoints.A else ch_2.extA

    virtual_nodes = []
    support_chain = None
    domain_interpolation(
        support_chain, ch1_border, ch2_border, endpoint, ch_1, virtual_nodes
    )

    return virtual_nodes


def regularity_of_the_derivative(
    state: Any,
    ch_j: Chain,
    ch_k: Chain,
    endpoint_j: EndPoints,
    node_list: List[Node],
    ch_j_nodes: List[Node],
    ch_k_nodes: List[Node],
    th_deriv: float = 1.0,
    derivative_from_center: bool = False,
) -> bool:
    """
    Checks the regularity of the derivative for the virtual nodes generated between two chains.

    Args:
        state (Any): State object for debugging purposes.
        ch_j (Chain): Source chain to be connected.
        ch_k (Chain): Destination chain to be connected.
        endpoint_j (EndPoints): Endpoint of ch_j to be connected.
        node_list (List[Node]): All the nodes involved in the connection, including the virtual nodes.
        ch_j_nodes (List[Node]): Nodes of ch_j.
        ch_k_nodes (List[Node]): Nodes of ch_k.
        th_deriv (float): Derivative threshold.
        derivative_from_center (bool): If True, regenerate virtual nodes interpolating from the center of ch_i.

    Returns:
        bool: True if the derivative is regular, False otherwise.
    """
    if derivative_from_center:
        new_list = []
        virtual_nodes = generate_virtual_nodes_without_support_chain(
            ch_j, ch_k, endpoint_j
        )
        angles = [n.angle for n in virtual_nodes]
        for node in node_list:
            if node.angle not in angles:
                new_list.append(node)
            else:
                new_list.append(get_node_from_list_by_angle(virtual_nodes, node.angle))
        node_list = new_list

    RegularDeriv = regularity_of_the_derivative_condition(
        state, ch_j.nr, node_list, ch_j_nodes, ch_k_nodes, endpoint_j, th_deriv=th_deriv
    )

    return RegularDeriv


def radial_tolerance_for_connecting_chains(
    state: Any, th_radial_tolerance: float, endpoints_radial: List[float]
) -> bool:
    """
    Checks maximum radial distance allowed to connect chains.

    Args:
        state (Any): State object for debugging purposes.
        th_radial_tolerance (float): Radial tolerance threshold.
        endpoints_radial (List[float]): Radial distances between endpoints and support chain.

    Returns:
        bool: True if radial distance is within tolerance, False otherwise.
    """
    delta_ri = endpoints_radial[0]
    delta_ri_plus_i = endpoints_radial[1]
    inf_delta_ri = delta_ri * (1 - th_radial_tolerance)
    sup_delta_ri = delta_ri * (1 + th_radial_tolerance)
    RadialTol = inf_delta_ri <= delta_ri_plus_i <= sup_delta_ri

    if state is not None and getattr(state, "debug", False):
        plt.figure()
        plt.axvline(x=delta_ri, color="b", label="delta_ri")
        plt.axvline(x=delta_ri_plus_i, color="r", label="delta_ri_plus_i")
        plt.axvline(x=inf_delta_ri, color="k", label="inf radial")
        plt.axvline(x=sup_delta_ri, color="k", label="sup radial")
        plt.title(f"{RadialTol}: Th {th_radial_tolerance}.")
        plt.savefig(f"{str(state.path)}/{state.counter}_max_radial_condition.png")
        plt.close()
        state.counter += 1

    return RadialTol


def similar_radial_distances_of_nodes_in_both_chains(
    state: Any, distribution_th: float, set_j: List[float], set_k: List[float]
) -> Tuple[bool, float]:
    """
    Checks if the radial distances of the nodes in both chains are similar via distribution of the radial distances.

    Args:
        state (Any): State object for debugging purposes.
        distribution_th (float): Size of the distributions to check if they are similar.
        set_j (List[float]): Radial distances between nodes of the first chain and the support chain.
        set_k (List[float]): Radial distances between nodes of the second chain and the support chain.

    Returns:
        Tuple[bool, float]: (Similarity result, distance between the mean of the distributions).
    """
    mean_j = np.mean(set_j)
    sigma_j = np.std(set_j)
    inf_range_j = mean_j - distribution_th * sigma_j
    sup_range_j = mean_j + distribution_th * sigma_j

    mean_k = np.mean(set_k)
    std_k = np.std(set_k)
    inf_range_k = mean_k - distribution_th * std_k
    sup_range_k = mean_k + distribution_th * std_k

    SimilarRadialDist = inf_range_k <= sup_range_j and inf_range_j <= sup_range_k

    if state is not None and getattr(state, "debug", False):
        plt.figure()
        plt.hist(set_j, bins=10, alpha=0.3, color="r", label="src radial")
        plt.hist(set_k, bins=10, alpha=0.3, color="b", label="dst radial")
        plt.axvline(x=inf_range_j, color="r", label="inf src")
        plt.axvline(x=sup_range_j, color="r", label="sup src")
        plt.axvline(x=inf_range_k, color="b", label="inf dst")
        plt.axvline(x=sup_range_k, color="b", label="sup dst")
        plt.legend()
        plt.title(f"{SimilarRadialDist}: Th {distribution_th}.")
        plt.savefig(f"{str(state.path)}/{state.counter}_distribution_condition.png")
        plt.close()
        state.counter += 1

    return SimilarRadialDist, np.abs(mean_j - mean_k)


def similarity_conditions(
    state: Any,
    th_radial_tolerance: float,
    th_distribution_size: float,
    th_regular_derivative: float,
    derivative_from_center: bool,
    ch_i: Chain,
    ch_j: Chain,
    candidate_chain: Chain,
    endpoint: EndPoints,
    check_overlapping: bool = True,
    chain_list: Optional[List[Chain]] = None,
) -> Tuple[bool, float]:
    """
    Checks the similarity conditions defined in Equation 6 in the paper.

    Args:
        state (Any): State object for debugging purposes.
        th_radial_tolerance (float): Radial tolerance threshold.
        th_distribution_size (float): Distribution size threshold.
        th_regular_derivative (float): Regular derivative threshold.
        derivative_from_center (bool): If True, regenerate virtual nodes from the center.
        ch_i (Chain): Support chain.
        ch_j (Chain): Chain j to connect.
        candidate_chain (Chain): Candidate chain to connect.
        endpoint (EndPoints): Endpoint of the source chain.
        check_overlapping (bool): Whether to check for overlapping chains.
        chain_list (Optional[List[Chain]]): List of chains to check for overlapping.

    Returns:
        Tuple[bool, float]: (Similarity condition result, distance between radial distributions).
    """
    neighbourhood = ChainNeighbourhood(ch_j, candidate_chain, ch_i, endpoint)

    if state is not None and getattr(state, "debug", False):
        visualize_selected_ch_and_chains_over_image_(
            [ch_i, ch_j, candidate_chain],
            state.l_ch_s,
            img=state.img,
            filename=f"{state.path}/{state.counter}_radial_conditions_{endpoint}.png",
        )
        state.counter += 1
        neighbourhood.draw_neighbourhood(
            f"{state.path}/{state.counter}_radials_conditions_neighbourdhood_{ch_j.label_id}_{candidate_chain.label_id}.png"
        )
        state.counter += 1

    if len(neighbourhood.set_i) <= 1 or len(neighbourhood.set_k) <= 1:
        return False, -1

    # 1. Radial tolerance for connecting chains
    RadialTol = radial_tolerance_for_connecting_chains(
        state, th_radial_tolerance, neighbourhood.radial_distance_endpoints_to_support
    )

    # 2. Similar radial distances of nodes in both chains
    SimilarRadialDist, distribution_distance = (
        similar_radial_distances_of_nodes_in_both_chains(
            state, th_distribution_size, neighbourhood.set_i, neighbourhood.set_k
        )
    )

    check_pass = RadialTol or SimilarRadialDist
    if not check_pass:
        return False, distribution_distance

    # 3. Derivative condition
    RegularDeriv = regularity_of_the_derivative(
        state,
        ch_j,
        candidate_chain,
        endpoint,
        neighbourhood.neighbourhood_nodes,
        ch_j_nodes=neighbourhood.src_chain_nodes,
        ch_k_nodes=neighbourhood.dst_chain_nodes,
        th_deriv=th_regular_derivative,
        derivative_from_center=derivative_from_center,
    )

    if not RegularDeriv:
        return False, distribution_distance

    # 4.0 Check there is no chains in region
    if check_overlapping:
        exist_chain = exist_chain_overlapping(
            state.l_ch_s if chain_list is None else chain_list,
            neighbourhood.endpoint_and_virtual_nodes,
            ch_j,
            candidate_chain,
            endpoint,
            ch_i,
        )
        if exist_chain:
            return False, distribution_distance

    return True, distribution_distance


def exist_chain_in_band_logic(
    chain_list: List[Chain], band_info: VirtualBandGenerator
) -> List[Chain]:
    """
    Checks for chains within the band.

    Args:
        chain_list (List[Chain]): List of chains to check.
        band_info (InfoVirtualBand): Band information.

    Returns:
        List[Chain]: List of chains found in the band.
    """
    chain_of_interest = [band_info.ch_k, band_info.ch_j]
    if band_info.ch_i is not None:
        chain_of_interest.append(band_info.ch_i)

    fist_chain_in_region = next(
        (
            chain
            for chain in chain_list
            if chain not in chain_of_interest and band_info.is_chain_in_band(chain)
        ),
        None,
    )

    return [fist_chain_in_region] if fist_chain_in_region is not None else []


def exist_chain_overlapping(
    l_ch_s: List[Chain],
    l_nodes: List[Node],
    ch_j: Chain,
    ch_k: Chain,
    endpoint_type: EndPoints,
    ch_i: Chain,
) -> bool:
    """
    Checks if there is a chain in the area within the band.

    Algorithm 11 in the supplementary material.

    Args:
        l_ch_s (List[Chain]): Full chains list.
        l_nodes (List[Node]): Both endpoints and virtual nodes.
        ch_j (Chain): Source chain.
        ch_k (Chain): Destination chain.
        endpoint_type (EndPoints): Endpoint type of ch_j.
        ch_i (Chain): Support chain.

    Returns:
        bool: True if a chain exists in the band, False otherwise.
    """
    info_band = VirtualBandGenerator(l_nodes, ch_j, ch_k, endpoint_type, ch_i)
    l_chains_in_band = exist_chain_in_band_logic(l_ch_s, info_band)
    exist_chain = len(l_chains_in_band) > 0

    return exist_chain



from typing import Any


class Node:
    def __init__(
        self, x: float, y: float, chain_id: int, radial_distance: float, angle: float
    ) -> None:
        """
        Node representing a point with coordinates, angle, and radial distance.

        Args:
            x (float): X-coordinate.
            y (float): Y-coordinate.
            chain_id (int): ID of the chain this node belongs to.
            radial_distance (float): Radial distance from center.
            angle (float): Angle in degrees.

        Returns:
            None
        """
        self.x = x
        self.y = y
        self.chain_id = chain_id
        self.radial_distance = radial_distance
        self.angle = angle

    def __repr__(self) -> str:
        """
        Representation of the node.

        Args:
            None

        Returns:
            str: Representation of the node
        """
        return f"({self.x},{self.y}) ang:{self.angle} radio:{self.radial_distance:0.2f} cad.id {self.chain_id}\n"

    def __str__(self) -> str:
        """
        String representation of the node.

        Args:
            None

        Returns:
            str: String representation of the node
        """
        return f"({self.x},{self.y}) ang:{self.angle} radio:{self.radial_distance:0.2f} id {self.chain_id}"

    def __eq__(self, other: Any) -> bool:
        """
        Compare two nodes.

        Args:
            other (Any): Node to compare

        Returns:
            bool: True if the nodes are equal, False otherwise
        """
        return (
            isinstance(other, Node)
            and self.x == other.x
            and self.y == other.y
            and self.angle == other.angle
        )



import os
from pathlib import Path
from typing import Dict, List, Tuple
from venv import logger

import cv2
import numpy as np
import pandas as pd
from importlib import resources
import platform

from ..config import config


def get_architecture() -> str:
    machine = platform.machine().lower()
    if machine.startswith("arm") or "aarch" in machine:
        return "arm64"
    elif machine in ("amd64", "x86_64"):
        return "x86_64"
    else:
        return "unknown"


def get_platform() -> str:
    system = platform.system().lower()
    if system == "linux":
        return "linux"
    elif system == "darwin":
        return "macos"
    elif system == "windows":
        return "windows"
    else:
        return "Unknown"


def get_devernay_path() -> str:
    """Get the absolute path to the appropriate devernay binary based on the platform."""
    base_package = "treediskrings.externals"

    # Determine subdirectory and binary name based on platform
    platform = get_platform()
    architecture = get_architecture()

    if platform == "linux":
        if architecture == "x86_64":
            subdir = "linux-x86_64"
            binary_name = "devernay.out"
        elif architecture == "arm64":
            subdir = "linux-arm"
            binary_name = "devernay.out"
        else:
            raise ValueError(f"Unsupported architecture: {architecture}")
    elif platform == "macos":
        if architecture == "x86_64":
            NotImplementedError(f"macOS x86_64 is not supported")
        elif architecture == "arm64":
            subdir = "macos-arm64"
            binary_name = "devernay.out"
        else:
            raise ValueError(f"Unsupported architecture: {architecture}")
    elif platform == "windows":
        raise NotImplementedError("Windows is not supported")
    else:
        raise ValueError(f"Unsupported platform: {platform}")

    # Construct the path to the binary using the resources API
    binary_path = resources.files(base_package) / subdir / binary_name
    binary_path = str(binary_path)

    # Ensure the file exists
    if not os.path.isfile(binary_path):
        raise FileNotFoundError(f"Binary not found at {binary_path}")

    # Make sure it's executable (skip on Windows)
    if platform != "windows":
        os.chmod(binary_path, 0o755)

    return binary_path


def load_curves(output_txt: Path) -> np.ndarray:
    """
    Loads curves from a text file.

    Args:
        output_txt (str): Path to the output text file containing curves.

    Returns:
        np.ndarray: Array of curves loaded from the file.
    """
    curves_list = pd.read_csv(output_txt, delimiter=" ", header=None).values
    return curves_list


def convert_image_to_pgm(img_pre: np.ndarray) -> Path:
    """
    Converts an image to PGM format and saves it.

    Args:
        img_pre (np.ndarray): Preprocessed image array.

    Returns:
        str: Path to the saved image file.
    """
    image_path = Path(config.output_dir) / "test.pgm"
    cv2.imwrite(str(image_path), img_pre)

    return image_path


def delete_files(files: List[Path]) -> None:
    """
    Deletes a list of files.

    Args:
        files (List[Path]): List of file paths to delete.
    """
    for file in files:
        os.system(f"rm {file}")


def gradient_load(
    img: np.ndarray, gx_path: str, gy_path: str
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Loads gradient images from files.

    Args:
        img (np.ndarray): Original image array for size reference.
        gx_path (str): Path to the gradient X file.
        gy_path (str): Path to the gradient Y file.

    Returns:
        Tuple[np.ndarray, np.ndarray]: Gradient images Gx and Gy.
    """
    Gx = np.zeros_like(img, dtype=float)
    Gy = np.zeros_like(img, dtype=float)
    Gx[1:-1, 1:-1] = pd.read_csv(gx_path, delimiter=" ", header=None).values.T
    Gy[1:-1, 1:-1] = pd.read_csv(gy_path, delimiter=" ", header=None).values.T

    return Gx, Gy


def execute_command(
    image_path: Path, sigma: float, low: float, high: float
) -> Tuple[Path, Path, Path]:
    """
    Executes the Devernay edge detector command.

    Args:
        image_path (Path): Path to the image file.
        sigma (float): Gaussian filter sigma value.
        low (float): Low threshold for edge detection.
        high (float): High threshold for edge detection.

    Returns:
        Tuple[Path, Path, Path]: Paths to the gradient X, gradient Y, and output text files.
    """
    output_txt = Path(config.output_dir) / "output.txt"
    gx_path = Path(config.output_dir) / "gx.txt"
    gy_path = Path(config.output_dir) / "gy.txt"
    devernay_path = get_devernay_path()
    command = (
        f"{devernay_path} {image_path} -s {sigma} -l {low} -h {high} "
        f"-t {output_txt} -x {gx_path} -y {gy_path}"
    )
    os.system(command)
    return gx_path, gy_path, output_txt


def canny_deverney_edge_detector(
    img_pre: np.ndarray, sigma: float, low: float, high: float
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Canny edge detector module using the Canny/Devernay algorithm.
    Downloaded from https://doi.org/10.5201/ipol.2017.216

    Args:
        img_pre (np.ndarray): Preprocessed image.
        sigma (float): Gaussian filter sigma value for edge detection.
        low (float): Low gradient threshold.
        high (float): High gradient threshold.

     Returns:
        Tuple[np.ndarray, np.ndarray, np.ndarray]:
            A tuple containing:
            - m_ch_e (np.ndarray): Devernay curves.
            - Gx (np.ndarray): Gradient image in the x direction.
            - Gy (np.ndarray): Gradient image in the y direction.
    """
    im_path = convert_image_to_pgm(img_pre)
    gx_path, gy_path, output_txt = execute_command(im_path, sigma, low, high)
    Gx, Gy = gradient_load(img_pre, str(gx_path), str(gy_path))
    m_ch_e = load_curves(output_txt)
    delete_files([output_txt, im_path, gx_path, gy_path])

    return m_ch_e, Gx, Gy



from typing import List

from ..geometry.chain import Chain


class ChainsBag:
    def __init__(self, inward_chain_set: List[Chain]) -> None:
        """
        Iterate over chains in a region.

        Args:
            inward_chain_set: list of chains in a region

        Returns:
            None
        """
        self.chain_set = inward_chain_set
        self.chains_id_already_selected = []

    def get_next_chain(self) -> Chain:
        """
        Get the next chain in the region.

        Returns:
            Chain: the next chain
        """
        next = None

        for chain in self.chain_set:
            if chain.id not in self.chains_id_already_selected:
                next = chain
                self.chains_id_already_selected.append(next.id)
                break

        return next



from shapely.geometry import Point, LineString
from typing import List, Optional, Tuple

from ..geometry.ring import Ring
from ..geometry.chain import Chain
from ..geometry.geometry_utils import visualize_selected_ch_and_chains_over_image_


class ChainContext:
    def __init__(
        self,
        l_ch_c: List[Chain],
        idx_start: Optional[int],
        save_path: Optional[str] = None,
        img: Optional[str] = None,
        debug: bool = True,
    ):
        """
        Initialize the ChainContext object.

        Args:
            l_ch_c (List[Chain]): List of chains
            idx_start (Optional[int]): Starting index
            save_path (Optional[str], optional): Path to save images. Defaults to None.
            img (Optional[str], optional): Image path. Defaults to None.
            debug (bool, optional): Debug flag. Defaults to True.
        """
        self.l_within_chains = []
        self.neighbourhood_size = None
        self.debug = debug
        self.save_path = save_path
        self.img = img
        self.completed_chains = [cad for cad in l_ch_c if cad.size >= cad.nr]
        self.completed_chains, self.poly_completed_chains = (
            self._from_completed_chain_to_poly(self.completed_chains)
        )

        self.uncompleted_chains = [cad for cad in l_ch_c if cad.size < cad.nr]
        self.uncompleted_chains_poly = self._from_uncompleted_chains_to_poly(
            self.uncompleted_chains
        )
        self.idx = 1 if idx_start is None else idx_start

    def get_inward_outward_ring(
        self, idx: int
    ) -> Tuple[Optional[Ring], Optional[Ring]]:
        """
        Get the inward and outward rings based on the index.

        Args:
            idx (int): Index to get the rings

        Returns:
            Tuple[Optional[Ring], Optional[Ring]]: Tuple of inward and outward rings
        """
        self.neighbourhood_size = 45
        outward_ring = None
        inward_ring = None

        if len(self.poly_completed_chains) > idx > 0:
            inward_ring = self.poly_completed_chains[idx - 1]
            outward_ring = self.poly_completed_chains[idx]

        return inward_ring, outward_ring

    def update(self) -> None:
        """
        Update the context. The context is updated when the algorithm is executed over a new region.
        """
        from ..processing.postprocessing import (
            search_shapely_inward_chain,
            from_shapely_to_chain,
        )

        inward_poly_ring, outward_poly_ring = self.get_inward_outward_ring(self.idx)
        shapely_inward_chain_subset = search_shapely_inward_chain(
            self.uncompleted_chains_poly, outward_poly_ring, inward_poly_ring
        )
        self.l_within_chains = from_shapely_to_chain(
            self.uncompleted_chains_poly,
            self.uncompleted_chains,
            shapely_inward_chain_subset,
        )

        self.inward_ring, self.outward_ring = self._from_shapely_ring_to_chain(
            inward_poly_ring, outward_poly_ring
        )

    def exit(self) -> bool:
        """
        Increment the index and check if it exceeds the length of completed chains.

        Returns:
            bool: True if index exceeds the length of completed chains, otherwise False
        """
        self.idx += 1
        if self.idx >= len(self.completed_chains):
            return True

        return False

    def drawing(self, iteration: int) -> None:
        """
        Visualize and save the selected chains and rings over the image.

        Args:
            iteration (int): Iteration number for the filename

        Returns:
            None
        """
        visualize_selected_ch_and_chains_over_image_(
            self.l_within_chains
            + [
                chain
                for chain in [self.inward_ring, self.outward_ring]
                if chain is not None
            ],
            [],
            img=self.img,
            filename=f"{self.save_path}/{iteration}_0.png",
        )

    def _from_shapely_ring_to_chain(
        self, poly_ring_inward: Optional[Ring], poly_ring_outward: Optional[Ring]
    ) -> Tuple[Optional[Ring], Optional[Ring]]:
        """
        Convert shapely rings to chain rings.

        Args:
            poly_ring_inward (Optional[Ring]): Inward shapely ring
            poly_ring_outward (Optional[Ring]): Outward shapely ring

        Returns:
            Tuple[Optional[Ring], Optional[Ring]]: Tuple of inward and outward chain rings
        """
        inward_chain_ring = None
        outward_chain_ring = None

        if poly_ring_inward is not None:
            inward_chain_ring = self.completed_chains[
                self.poly_completed_chains.index(poly_ring_inward)
            ]

        if poly_ring_outward is not None:
            outward_chain_ring = self.completed_chains[
                self.poly_completed_chains.index(poly_ring_outward)
            ]

        return inward_chain_ring, outward_chain_ring

    def sort_list_by_index_array(self, indexes: List[int], list: List) -> List:
        """
        Sort a list based on the given indexes.

        Args:
            indexes (List[int]): List of indexes
            list_position (List): List to be sorted

        Returns:
            List: Sorted list
        """
        Z = [list[position] for position in indexes]

        return Z

    def sort_shapely_list_and_chain_list(
        self, cadena_list: List[Chain], shapely_list: List[Ring]
    ) -> Tuple[List[Chain], List[Ring]]:
        """
        Sort shapely list and chain list based on the area of shapely geometries.

        Args:
            cadena_list (List[Chain]): List of chains
            shapely_list (List[Ring]): List of shapely geometries

        Returns:
            Tuple[List, List]: Tuple of sorted chain list and shapely list
        """
        idx_sort = [
            i[0]
            for i in sorted(enumerate(shapely_list), key=lambda x: x[1].geometry.area)
        ]
        cadena_list = self.sort_list_by_index_array(idx_sort, cadena_list)
        shapely_list = self.sort_list_by_index_array(idx_sort, shapely_list)

        return cadena_list, shapely_list

    def _from_completed_chain_to_poly(
        self, completed_chain: List[Chain]
    ) -> Tuple[List[Chain], List[Ring]]:
        """
        Convert completed chains to shapely polygons.

        Args:
            completed_chain (List[Chain]): List of completed chains

        Returns:
            Tuple[List[Chain], List[Ring]]: Tuple of completed chains and shapely polygons
        """
        poly_completed_chains = []

        for chain in completed_chain:
            ring = Ring(chain, id=chain.id)
            poly_completed_chains.append(ring)

        completed_chain, poly_completed_chains = self.sort_shapely_list_and_chain_list(
            completed_chain, poly_completed_chains
        )

        return completed_chain, poly_completed_chains

    def _from_uncompleted_chains_to_poly(
        self, uncompleted_chain: List[Chain]
    ) -> List[LineString]:
        """
        Convert uncompleted chains to shapely LineStrings.

        Args:
            uncompleted_chain (List): List of uncompleted chains

        Returns:
            List[LineString]: List of shapely LineStrings
        """
        uncompleted_chain_shapely = []

        for chain in uncompleted_chain:
            lista_pts = [Point(punto.y, punto.x) for punto in chain.sort_dots()]
            uncompleted_chain_shapely.append(LineString(lista_pts))

        return uncompleted_chain_shapely



import numpy as np
from typing import List, Tuple, Optional

from ..geometry.geometry_utils import (
    copy_chain,
    angular_distance_between_chains,
    visualize_selected_ch_and_chains_over_image_,
    euclidean_distance_between_nodes,
    get_chain_from_list_by_id,
    minimum_euclidean_distance_between_chains_endpoints,
    get_chains_within_angle,
    get_closest_dots_to_angle_on_radial_direction_sorted_by_ascending_distance_to_center,
)
from ..geometry.node import Node
from ..geometry.angular_set import Set
from ..analysis.chain_system_manager import ChainSystemManager
from ..analysis.connect_parameters import ConnectParameters
from ..geometry.chain import Chain, TypeChains, EndPoints, ChainLocation
from ..geometry.node import Node
from ..analysis.interpolation_nodes import (
    compute_interpolation_domain,
    domain_interpolation,
)
from ..analysis.chain_analysis_tools import similarity_conditions
from ..config import config


def extract_border_chain_from_list(
    ch_s: List[Chain], nodes_s: List[Node]
) -> Tuple[Chain, List[Chain], List[Node]]:
    """
    Extract border chain from chain and nodes list.

    Args:
        ch_s (List[Chain]): Chain list.
        nodes_s (List[Node]): Node list.

    Returns:
        Tuple[Chain, List[Chain], List[Node]]: Border chain, chains without border, nodes without border.
    """
    ch_s_without_border = [chain for chain in ch_s if chain.type != TypeChains.border]
    border_chain = next(chain for chain in ch_s if (chain.type == TypeChains.border))
    nodes_s_without_border = [
        node for node in nodes_s if node.chain_id != border_chain.id
    ]

    return border_chain, ch_s_without_border, nodes_s_without_border


def copy_chains_and_nodes(ch_s: List[Chain]) -> Tuple[List[Chain], List[Node]]:
    """
    Copy chains and their nodes.

    Args:
        ch_s (List[Chain]): Chain list.

    Returns:
        Tuple[List[Chain], List[Node]]: Copied chains and their nodes.
    """
    nodes_s = []

    ch_s = [copy_chain(chain) for chain in ch_s]

    for chain in ch_s:
        nodes_s += chain.l_nodes

    return ch_s, nodes_s


def connect_chains(
    l_ch_s: List[Chain],
    img_pre: Optional[np.ndarray] = None,
) -> Tuple[List[Chain], List[Node]]:
    """
    Logic to connect chains. Same logic to connect chains is applied several times, smoothing restriction.
    Implements Algorithm 7 in the supplementary material.

    Args:
        l_ch_s (List[Chain]): Chain list.
        nr (int): Some integer parameter.
        debug (bool): Debug flag.
        img_pre (bool): Some boolean flag.
        output_dir (str): Output directory.

    Returns:
        Tuple[List[Chain], List[Node]]: Connected chains and their nodes.
    """
    # Copy ch_i and nodes
    l_ch_s, l_nodes_s = copy_chains_and_nodes(l_ch_s)
    # Paper Table 1 parameters initialization
    parameters = ConnectParameters(l_ch_s, l_nodes_s)

    m = compute_intersection_matrix(l_ch_s, l_nodes_s)

    # Iteration over the parameters, 9 iterations in total
    for i in range(parameters.iterations):
        # Get parameters for iteration i
        iteration_params = parameters.get_iteration_parameters(i)

        # debug_img_pre is a copy of the input image. It is used to visualize the results of the current iteration for debugging purposes.
        # Algorithm 2 in the paper
        l_ch_c, l_nodes_c, m = connect_chains_main_logic(
            m=m,
            nr=config.nr,
            debug_imgs=config.debug,
            img_pre=img_pre,
            save=f"{config.output_dir}/output_{i}_",
            **iteration_params,
        )

        # Update chain list and chain node list for next iteration
        parameters.update_list_for_next_iteration(l_ch_c, l_nodes_c)

    return l_ch_c, l_nodes_c


def update_pointer(
    ch_j: Chain, closest: Optional[Chain], l_candidates_chi: List[Chain]
) -> int:
    """
    Update the pointer for the chain.

    Args:
        ch_j (Chain): The current chain.
        closest (Optional[Chain]): The closest chain.
        l_candidates_chi (List[Chain]): List of candidate chains.

    Returns:
        int: The updated pointer.
    """
    ch_j_index = l_candidates_chi.index(ch_j)
    j_pointer = ch_j_index if closest is not None else ch_j_index + 1

    return j_pointer


def iterate_over_chains_list_and_complete_them_if_met_conditions(
    state: ChainSystemManager,
) -> Tuple[List[Chain], List[Node], np.ndarray]:
    """
    Iterate over the list of chains and complete them if conditions are met.

    Args:
        state (SystemStatus): The current system state.

    Returns:
        Tuple[List[Chain], List[Node], np.ndarray]: Updated chains, nodes, and intersection matrix.
    """
    for chain in state.l_ch_s:
        state.fill_chain_if_there_is_no_overlapping(chain)

    return state.l_ch_s, state.l_nodes_s, state.M


def debugging_chains(
    state: ChainSystemManager, chains_to_debug: List[Chain], filename: str
) -> None:
    """
    Debug the chains by visualizing them.

    Args:
        state (SystemStatus): The current system state.
        chains_to_debug (List[Chain]): List of chains to debug.
        filename (str): Filename to save the debug image.

    Returns:
        None
    """
    if state.debug:
        visualize_selected_ch_and_chains_over_image_(
            [ch for ch in chains_to_debug if ch is not None],
            state.l_ch_s,
            img=state.img,
            filename=filename,
        )
        state.counter += 1


def connect_chains_main_logic(
    m: np.ndarray,
    nr: int,
    l_ch_s: List[Chain],
    l_nodes_s: List[Node],
    th_radial_tolerance: float = 2,
    th_distribution_size: float = 2,
    th_regular_derivative: float = 1.5,
    neighbourhood_size: int = 22,
    derivative_from_center: bool = False,
    debug_imgs: bool = False,
    img_pre: Optional[np.ndarray] = None,
    save: Optional[str] = None,
) -> Tuple[List[Chain], List[Node], np.ndarray]:
    """
    Logic for connecting chains based on similarity conditions. Implements Algorithm 2 from the paper.

    Args:
        m (np.ndarray): Matrix of intersections between chains.
        nr (int): Number of rays.
        l_ch_s (List[Chain]): List of chains.
        l_nodes_s (List[Node]): List of nodes belonging to chains.
        th_radial_tolerance (float, optional): Threshold for radial tolerance. Defaults to 2.
        th_distribution_size (float, optional): Threshold for distribution size. Defaults to 2.
        th_regular_derivative (float, optional): Threshold for regular derivative. Defaults to 1.5.
        neighbourhood_size (int, optional): Size of neighbourhood in which we search for similar chains. Defaults to 22.
        derivative_from_center (bool, optional): If true, derivative is calculated from cy, otherwise from support chain. Defaults to False.
        debug_imgs (bool, optional): Debug parameter. Defaults to False.
        img_pre (Optional[np.ndarray], optional): Image for debug. Defaults to None.
        save (Optional[str], optional): Image save location. Debug only. Defaults to None.

    Returns:
        Tuple[List[Chain], List[Node], np.ndarray]: Nodes and chain list after connecting.
    """
    # Initialization of the state object. It contains all the information needed to connect chains.
    state = ChainSystemManager(
        l_ch_s,
        l_nodes_s,
        m,
        nr=nr,
        th_radial_tolerance=th_radial_tolerance,
        th_distribution_size=th_distribution_size,
        th_regular_derivative=th_regular_derivative,
        neighbourhood_size=neighbourhood_size,
        derivative_from_center=derivative_from_center,
        debug=debug_imgs,
        save=save,
        img=img_pre,
    )

    # State.continue_in_loop() check if current state is equal to the previous one. If it is, the algorithm stops.
    # If some chains have been connected in the current iteration, the algorithm continues one more iteration.
    # Additionaly, if nodes have been added to some chain, the algorithm continues one more iteration.
    while state.continue_in_loop():
        # Get next chain to be processed. Algorithm 11 in the paper.
        ch_i = state.get_next_chain()

        # Get chains in ch_i neighbourhood
        l_s_outward, l_s_inward = get_chains_in_and_out_wards(state.l_ch_s, ch_i)

        for location, l_candidates_chi in zip(
            [ChainLocation.inwards, ChainLocation.outwards],
            [l_s_inward, l_s_outward],
        ):
            j_pointer = 0
            while len(l_candidates_chi) > j_pointer:
                debugging_chains(
                    state,
                    [ch_i] + l_candidates_chi,
                    f"{state.path}/{state.counter}_0_{ch_i.label_id}_{location}.png",
                )
                ch_j = l_candidates_chi[j_pointer]

                debugging_chains(
                    state, [ch_i, ch_j], f"{state.path}/{state.counter}_1.png"
                )
                l_no_intersection_j = get_non_intersection_chains(
                    state.M, l_candidates_chi, ch_j
                )

                # Algorithm 13 in the supplementary material
                ch_k_b = get_closest_chain_logic(
                    state,
                    ch_j,
                    l_candidates_chi,
                    l_no_intersection_j,
                    ch_i,
                    location,
                    EndPoints.B,
                )
                debugging_chains(
                    state, [ch_i, ch_j, ch_k_b], f"{state.path}/{state.counter}_2.png"
                )

                # Algorithm 13 in the supplementary material
                ch_k_a = get_closest_chain_logic(
                    state,
                    ch_j,
                    l_candidates_chi,
                    l_no_intersection_j,
                    ch_i,
                    location,
                    EndPoints.A,
                )
                debugging_chains(
                    state, [ch_i, ch_j, ch_k_a], f"{state.path}/{state.counter}_3.png"
                )

                ch_k, endpoint = select_closest_chain(ch_j, ch_k_a, ch_k_b)
                debugging_chains(
                    state, [ch_i, ch_j, ch_k], f"{state.path}/{state.counter}_4.png"
                )

                # Algorithm 14 in the paper
                connect_two_chains(state, ch_j, ch_k, l_candidates_chi, endpoint, ch_i)
                debugging_chains(
                    state, [ch_i, ch_j], f"{state.path}/{state.counter}_5.png"
                )

                j_pointer = update_pointer(ch_j, ch_k, l_candidates_chi)

        # Implementing the logic of Algorithm 10
        state.update_system_status(ch_i, l_s_outward, l_s_inward)

    l_ch_c, l_nodes_c, intersection_matrix = (
        iterate_over_chains_list_and_complete_them_if_met_conditions(state)
    )
    debugging_chains(state, l_ch_c, f"{state.path}/{state.counter}.png")

    return l_ch_c, l_nodes_c, intersection_matrix


def intersection_chains(
    m: np.ndarray, candidate_chain: Chain, l_sorted_chains_in_neighbourhood: List[Chain]
) -> List[Chain]:
    """
    Find chains that intersect with the candidate chain.

    Args:
        m (np.ndarray): Intersection matrix.
        candidate_chain (Chain): The candidate chain.
        l_sorted_chains_in_neighbourhood (List[Chain]): Chains in the neighbourhood sorted by angular distance.

    Returns:
        List[Chain]: List of chains that intersect with the candidate chain.
    """
    inter_next_chain = np.where(m[candidate_chain.id] == 1)[0]
    l_intersections_candidate = [
        set.cad
        for set in l_sorted_chains_in_neighbourhood
        if set.cad.id in inter_next_chain and candidate_chain.id != set.cad.id
    ]

    return l_intersections_candidate


def get_all_chain_in_subset_that_satisfy_condition(
    state: ChainSystemManager,
    ch_j: Chain,
    ch_i: Chain,
    endpoint: int,
    radial_distance: float,
    candidate_chain: Chain,
    l_intersections_candidate: List[Chain],
) -> List[Set]:
    """
    Get all chains in the subset that satisfy the connectivity goodness condition.

    Args:
        state (SystemStatus): The current system state.
        ch_j (Chain): The current chain.
        ch_i (Chain): The support chain.
        endpoint (int): The endpoint of ch_j.
        radial_distance (float): Radial distance between ch_j and candidate chain.
        candidate_chain (Chain): The candidate chain.
        l_intersections_candidate (List[Chain]): List of chains that intersect with the candidate chain.

    Returns:
        List[Set]: List of sets that satisfy the connectivity goodness condition.
    """
    l_intersection_candidate_set = [Set(radial_distance, candidate_chain)]

    for chain_inter in l_intersections_candidate:
        pass_control, radial_distance = connectivity_goodness_condition(
            state, ch_j, chain_inter, ch_i, endpoint
        )
        if pass_control:
            l_intersection_candidate_set.append(Set(radial_distance, chain_inter))

    return l_intersection_candidate_set


def get_the_closest_chain_by_radial_distance_that_does_not_intersect(
    state: ChainSystemManager,
    ch_j: Chain,
    ch_i: Chain,
    endpoint: int,
    candidate_chain_radial_distance: float,
    candidate_chain: Chain,
    M: np.ndarray,
    l_sorted_chains_in_neighbourhood: List[Chain],
) -> Chain:
    """
    Implements Algorithm 14 in the supplementary material.

    Args:
        state (SystemStatus): Data structure with all the information of the system.
        ch_j (Chain): The current chain.
        ch_i (Chain): The support chain.
        endpoint (int): The endpoint of ch_j.
        candidate_chain_radial_distance (float): Radial distance between ch_j and candidate chain.
        candidate_chain (Chain): Angular closer chain to ch_j.
        M (np.ndarray): Intersection matrix.
        l_sorted_chains_in_neighbourhood (List[Chain]): Chains in ch_j endpoint neighbourhood sorted by angular distance.

    Returns:
        Chain: Closest chain to ch_j that satisfies connectivity goodness conditions.
    """
    # Get all the chains that intersect to candidate_chain
    l_intersections_candidate = intersection_chains(
        M, candidate_chain, l_sorted_chains_in_neighbourhood
    )

    # Get all the chains that intersect to candidate_chain and satisfy connectivity_goodness_condition with ch_j
    l_intersections_candidate_set = get_all_chain_in_subset_that_satisfy_condition(
        state,
        ch_j,
        ch_i,
        endpoint,
        candidate_chain_radial_distance,
        candidate_chain,
        l_intersections_candidate,
    )
    # Sort them by proximity to ch_j
    l_intersections_candidate_set.sort(key=lambda x: x.distance)

    # Return ch_k ch_i
    ch_k = l_intersections_candidate_set[0].cad

    return ch_k


def get_closest_chain(
    state: ChainSystemManager,
    ch_j: Chain,
    l_no_intersection_j: List[Chain],
    ch_i: Chain,
    location: int,
    endpoint: int,
    m: np.ndarray,
) -> Optional[Chain]:
    """
    Search for the closest chain to ch_j that does not intersect with ch_j and met conditions.
    Implements Algorithm 3 from the paper.

    Args:
        state (SystemStatus): System status instance.
        ch_j (Chain): Source chain.
        l_no_intersection_j (List[Chain]): List of chains that do not intersect with ch_j.
        ch_i (Chain): Support chain of ch_j.
        location (int): Inward or outward ch_j location regarding ch_i.
        endpoint (int): ch_j endpoint.
        m (np.ndarray): Intersection matrix.

    Returns:
        Optional[Chain]: The closest chain to ch_j.
    """
    # Sort chains by proximity
    neighbourhood_size = state.neighbourhood_size
    l_sorted_chains_in_neighbourhood = get_chains_in_neighbourhood(
        neighbourhood_size, l_no_intersection_j, ch_j, ch_i, endpoint, location
    )

    next_id = 0
    ch_k = None

    # Search for closest chain to ch_i
    lenght_chains = len(l_sorted_chains_in_neighbourhood)
    while next_id < lenght_chains:
        candidate_chain = l_sorted_chains_in_neighbourhood[next_id].cad

        # Algorithm 4 from paper.
        pass_control, radial_distance = connectivity_goodness_condition(
            state, ch_j, candidate_chain, ch_i, endpoint
        )

        if pass_control:
            # Check that do not exist other chains that intersect next ch_i that is radially ch_k to ch_j
            # Get chains that intersect next ch_i. Algorithm 14 in the supplementary material.
            ch_k = get_the_closest_chain_by_radial_distance_that_does_not_intersect(
                state,
                ch_j,
                ch_i,
                endpoint,
                radial_distance,
                candidate_chain,
                m,
                l_sorted_chains_in_neighbourhood,
            )

            break

        next_id += 1

    return ch_k


def get_closest_chain_logic(
    state: ChainSystemManager,
    ch_j: Chain,
    l_candidates_chi: List[Chain],
    l_no_intersection_j: List[Chain],
    ch_i: Chain,
    location: int,
    endpoint: int,
) -> Optional[Chain]:
    """
    Get the ch_k chain that met condition if it is symmetric. If it is not symmetric return None.
    Implements Algorithm 13 in the supplementary material.

    Args:
        state (SystemStatus): System status instance.
        ch_j (Chain): Chain that is going to be connected to another chain.
        l_candidates_chi (List[Chain]): List of chains that can be candidates to be connected to ch_j.
        l_no_intersection_j (List[Chain]): List of chains that do not intersect with ch_j.
        ch_i (Chain): Chain that supports ch_j.
        location (int): Location of ch_j regarding support chain (inward/outward).
        endpoint (int): Endpoint of ch_j that is going to be connected.

    Returns:
        Optional[Chain]: Closest chain, ch_k, to ch_j that met condition.
    """
    # Algorithm 3 in the paper
    ch_k = get_closest_chain(
        state, ch_j, l_no_intersection_j, ch_i, location, endpoint, state.M
    )

    if ch_k is None:
        return ch_k

    l_no_intersection_k = get_non_intersection_chains(state.M, l_candidates_chi, ch_k)

    endpoint_k = EndPoints.A if endpoint == EndPoints.B else EndPoints.B

    symmetric_chain = get_closest_chain(
        state, ch_k, l_no_intersection_k, ch_i, location, endpoint_k, state.M
    )

    ch_k = None if symmetric_chain != ch_j else ch_k
    if ch_k is not None and (ch_k.size + ch_j.size) > ch_k.nr:
        ch_k = None

    return ch_k


def move_nodes_from_one_chain_to_another(ch_j: Chain, ch_k: Chain) -> bool:
    """
    Move nodes from one chain to another.

    Args:
        ch_j (Chain): Destination chain.
        ch_k (Chain): Source chain.

    Returns:
        bool: Whether the border has changed.
    """
    for node in ch_k.l_nodes:
        node.chain_id = ch_j.id

    change_border = ch_j.add_nodes_list(ch_k.l_nodes)

    return change_border


def generate_new_nodes(
    state: ChainSystemManager, ch_j: Chain, ch_k: Chain, endpoint: int, ch_i: Chain
) -> None:
    """
    Generate new nodes between chains.

    Args:
        state (SystemStatus): System status instance.
        ch_j (Chain): Destination chain.
        ch_k (Chain): Source chain.
        endpoint (int): Endpoint of ch_j.
        ch_i (Chain): Support chain.
    """
    l_new_nodes = []

    ch_j_endpoint = ch_j.extA if endpoint == EndPoints.A else ch_j.extB
    ch_k_endpoint = ch_k.extB if endpoint == EndPoints.A else ch_k.extA

    domain_interpolation(
        ch_i, ch_j_endpoint, ch_k_endpoint, endpoint, ch_j, l_new_nodes
    )
    state.add_nodes_list_to_system(ch_j, l_new_nodes)

    return


def updating_chain_nodes(state: ChainSystemManager, ch_j: Chain, ch_k: Chain) -> None:
    """
    Update chain nodes after moving nodes from one chain to another.

    Args:
        state (SystemStatus): System status instance.
        ch_j (Chain): Destination chain.
        ch_k (Chain): Source chain.
    """
    change_border = move_nodes_from_one_chain_to_another(ch_j, ch_k)
    if change_border:
        state.update_chain_neighbourhood([ch_j])

    return


def delete_closest_chain(
    state: ChainSystemManager, ch_k: Chain, l_candidates_chi: List[Chain]
) -> None:
    """
    Delete the closest chain from the system.

    Args:
        state (SystemStatus): System status instance.
        ch_k (Chain): Chain to be deleted.
        l_candidates_chi (List[Chain]): List of candidate chains.
    """
    cad_2_index = state.l_ch_s.index(ch_k)
    del state.l_ch_s[cad_2_index]
    id_connected_chain = l_candidates_chi.index(ch_k)
    del l_candidates_chi[id_connected_chain]

    return


def update_intersection_matrix(
    state: ChainSystemManager, ch_j: Chain, ch_k: Chain
) -> None:
    """
    Update the intersection matrix after deleting a chain.

    Args:
        state (SystemStatus): System status instance.
        ch_j (Chain): Remaining chain.
        ch_k (Chain): Deleted chain.
    """
    inter_cad_1 = state.M[ch_j.id]
    inter_cad_2 = state.M[ch_k.id]
    or_inter_cad1_cad2 = np.logical_or(inter_cad_1, inter_cad_2)
    state.M[ch_j.id] = or_inter_cad1_cad2
    state.M[:, ch_j.id] = or_inter_cad1_cad2
    state.M = np.delete(state.M, ch_k.id, 1)
    state.M = np.delete(state.M, ch_k.id, 0)

    return


def update_chains_ids(state: ChainSystemManager, ch_k: Chain) -> None:
    """
    Update chain IDs after deleting a chain.

    Args:
        state (SystemStatus): System status instance.
        ch_k (Chain): Deleted chain.
    """
    for ch_old in state.l_ch_s:
        if ch_old.id > ch_k.id:
            new_id = ch_old.id - 1
            ch_old.change_id(new_id)

    return


def connect_two_chains(
    state: ChainSystemManager,
    ch_j: Chain,
    ch_k: Chain,
    l_candidates_chi: List[Chain],
    endpoint: int,
    ch_i: Chain,
) -> None:
    """
    Algorithm 12 in the supplementary material. Connect chains ch_j and ch_k updating all the information about the system.

    Args:
        state (SystemStatus): Class object that contains all the information about the system.
        ch_j (Chain): Chain j to connect.
        ch_k (Chain): Chain k to connect.
        l_candidates_chi (List[Chain]): List of chains that have support chain ch_i.
        endpoint (int): Endpoint of ch_j that is going to be connected.
        ch_i (Chain): Support chain.

    Returns:
        None
    """
    if endpoint is None:
        return

    if ch_j == ch_k:
        return

    # Generate new dots
    generate_new_nodes(state, ch_j, ch_k, endpoint, ch_i)

    # move node from one ch_i to another
    updating_chain_nodes(state, ch_j, ch_k)

    # update chains
    update_chain_after_connect(state, ch_j, ch_k)

    # delete ch_k from  list l_candidate_chi  and state.l_ch_s
    delete_closest_chain(state, ch_k, l_candidates_chi)

    # update intersection matrix
    update_intersection_matrix(state, ch_j, ch_k)

    # update ch_i ids
    update_chains_ids(state, ch_k)

    return


def get_inward_and_outward_list_chains_via_pointers(
    l_ch_s: List[Chain], support_chain: Chain
) -> Tuple[List[Chain], List[Chain]]:
    """
    Get the inward and outward chains of ch_i.

    Args:
        l_ch_s (List[Chain]): List of chains.
        support_chain (Chain): Support chain, ch_i, to get the inward and outward chains.

    Returns:
        Tuple[List[Chain], List[Chain]]: Inward and outward list chains.
    """
    l_s_outward = []
    l_s_inward = []

    for ch_cand in l_ch_s:
        if ch_cand == support_chain:
            continue
        a_outward, b_outward, a_inward, b_inward = (
            ch_cand.A_outward,
            ch_cand.B_outward,
            ch_cand.A_inward,
            ch_cand.B_inward,
        )

        if (ch_cand not in l_s_outward) and (
            (a_inward is not None and support_chain is a_inward)
            or (b_inward is not None and support_chain is b_inward)
        ):
            l_s_outward.append(ch_cand)

        if (ch_cand not in l_s_inward) and (
            (a_outward is not None and support_chain is a_outward)
            or (b_outward is not None and support_chain is b_outward)
        ):
            l_s_inward.append(ch_cand)

    return l_s_outward, l_s_inward


def get_non_intersection_chains(
    M: np.ndarray, l_candidates_chi: List[Chain], ch_j: Chain
) -> List[Chain]:
    """
    Get the list of chains that do not intersect with ch_j.

    Args:
        M (np.ndarray): Intersection matrix.
        l_candidates_chi (List[Chain]): List of chains.
        ch_j (Chain): Chain j.

    Returns:
        List[Chain]: List of chains that do not intersect with ch_j.
    """
    id_inter = np.where(M[ch_j.id] == 1)[0]
    candidates_chi_non_chj_intersection = [
        cad for cad in l_candidates_chi if cad.id not in id_inter
    ]

    return candidates_chi_non_chj_intersection


def get_intersection_chains(
    M: np.ndarray, l_candidates_chi: List[Chain], ch_j: Chain
) -> List[Chain]:
    """
    Get the list of chains that intersect with ch_j.

    Args:
        M (np.ndarray): Intersection matrix.
        l_candidates_chi (List[Chain]): List of chains.
        ch_j (Chain): Chain j.

    Returns:
        List[Chain]: List of chains that intersect with ch_j.
    """
    id_inter = np.where(M[ch_j.id] == 1)[0]
    candidates_chi_non_chj_intersection = [
        cad for cad in l_candidates_chi if cad.id in id_inter
    ]

    return candidates_chi_non_chj_intersection


def remove_chains_if_present_at_both_groups(
    S_up: List[Chain], S_down: List[Chain]
) -> List[Chain]:
    """
    Remove chains if present in both groups.

    Args:
        S_up (List[Chain]): List of chains in the upward group.
        S_down (List[Chain]): List of chains in the downward group.

    Returns:
        List[Chain]: List of chains present in both groups.
    """
    up_down = [cad for cad in S_up if cad in S_down]
    for cad in up_down:
        S_up.remove(cad)

    return up_down


def get_chains_in_and_out_wards(
    l_ch_s: List[Chain], support_chain: Chain
) -> Tuple[List[Chain], List[Chain]]:
    """
    Get chains inwards and outwards from l_ch_s given support chain, ch_i.

    Args:
        l_ch_s (List[Chain]): List of chains.
        support_chain (Chain): Support chain, ch_i.

    Returns:
        Tuple[List[Chain], List[Chain]]: List of chains inwards and list of chains outwards.
    """
    l_s_outward, l_s_inward = get_inward_and_outward_list_chains_via_pointers(
        l_ch_s, support_chain
    )
    remove_chains_if_present_at_both_groups(l_s_outward, l_s_inward)

    return l_s_outward, l_s_inward


def select_closest_chain(
    chain: Chain, a_neighbour_chain: Optional[Chain], b_neighbour_chain: Optional[Chain]
) -> Tuple[Optional[Chain], Optional[int]]:
    """
    Select the closest chain to the given chain.

    Args:
        chain (Chain): The reference chain.
        a_neighbour_chain (Optional[Chain]): Neighbour chain at endpoint A.
        b_neighbour_chain (Optional[Chain]): Neighbour chain at endpoint B.

    Returns:
        Tuple[Optional[Chain], Optional[int]]: Closest chain and its endpoint.
    """
    if a_neighbour_chain is not None:
        d_a = distance_between_border(chain, a_neighbour_chain, EndPoints.A)
    else:
        d_a = -1

    if b_neighbour_chain is not None:
        d_b = distance_between_border(chain, b_neighbour_chain, EndPoints.B)
    else:
        d_b = -1

    if d_a == d_b == -1:
        closest_chain = None
        endpoint = None

    elif d_a >= d_b:
        closest_chain = a_neighbour_chain
        endpoint = EndPoints.A

    elif d_b > d_a:
        closest_chain = b_neighbour_chain
        endpoint = EndPoints.B

    else:
        raise

    return closest_chain, endpoint


def get_chains_in_neighbourhood(
    neighbourhood_size: float,
    l_no_intersection_j: List[Chain],
    ch_j: Chain,
    ch_i: Chain,
    endpoint: int,
    location: int,
) -> List[Set]:
    """
    Get all the chains in the neighbourhood of the chain ch_j included in the list no_intersection_j.

    Args:
        neighbourhood_size (float): Angular neighbourhood size.
        l_no_intersection_j (List[Chain]): List of chains that do not intersect with ch_j.
        ch_j (Chain): Chain j.
        ch_i (Chain): Support chain, ch_i.
        endpoint (int): ch_j endpoint.
        location (int): Inward or outward location.

    Returns:
        List[Set]: List of chains in the neighbourhood of ch_j.
    """
    l_chains_in_neighbourhood = []

    for cand_chain in l_no_intersection_j:
        angular_distance = angular_distance_between_chains(ch_j, cand_chain, endpoint)
        if angular_distance < neighbourhood_size and cand_chain.id != ch_j.id:
            l_chains_in_neighbourhood.append(Set(angular_distance, cand_chain))

    if endpoint == EndPoints.A and location == ChainLocation.inwards:
        l_chains_in_neighbourhood = [
            element
            for element in l_chains_in_neighbourhood
            if element.cad.B_outward == ch_i
        ]

    elif endpoint == EndPoints.A and location == ChainLocation.outwards:
        l_chains_in_neighbourhood = [
            element
            for element in l_chains_in_neighbourhood
            if element.cad.B_inward == ch_i
        ]
    elif endpoint == EndPoints.B and location == ChainLocation.inwards:
        l_chains_in_neighbourhood = [
            element
            for element in l_chains_in_neighbourhood
            if element.cad.A_outward == ch_i
        ]

    elif endpoint == EndPoints.B and location == ChainLocation.outwards:
        l_chains_in_neighbourhood = [
            element
            for element in l_chains_in_neighbourhood
            if element.cad.A_inward == ch_i
        ]

    l_sorted_chains_in_neighbourhood = sort_chains_in_neighbourhood(
        l_chains_in_neighbourhood, ch_j
    )

    return l_sorted_chains_in_neighbourhood


def sort_chains_in_neighbourhood(
    chains_in_neighbourhood: List[Set], ch_j: Chain
) -> List[Set]:
    """
    Sort chains by angular distance. Set of chains with same angular distance, are sorted by euclidean distance to ch_j.

    Args:
        chains_in_neighbourhood (List[Set]): List of Sets. A set elements is composed by a chain and a distance between support chain and ch_j.
        ch_j (Chain): Chain j.

    Returns:
        List[Set]: Sorted list of chains in the neighbourhood.
    """
    sorted_chains_in_neighbourhood = []

    unique_angular_distances = np.unique(
        [conj.distance for conj in chains_in_neighbourhood]
    )

    for d in unique_angular_distances:
        chains_same_angular_distance = [
            conj.cad for conj in chains_in_neighbourhood if conj.distance == d
        ]
        euclidean_distance_set = [
            Set(minimum_euclidean_distance_between_chains_endpoints(ch_d, ch_j), ch_d)
            for ch_d in chains_same_angular_distance
        ]
        euclidean_distance_set.sort(key=lambda x: x.distance)
        sorted_chains_in_neighbourhood += [
            Set(d, set.cad) for set in euclidean_distance_set
        ]

    return sorted_chains_in_neighbourhood


def check_endpoints(
    support_chain: Chain, ch_j: Chain, candidate_chain: Chain, endpoint: int
) -> bool:
    """
    Check if the endpoints of the chain ch_j are in the interpolation domain of the support chain, ch_i.

    Args:
        support_chain (Chain): Support chain of ch_j and candidate_chain.
        ch_j (Chain): Chain j.
        candidate_chain (Chain): Candidate chain.
        endpoint (int): ch_j endpoint.

    Returns:
        bool: True if the endpoints are in the interpolation domain, False otherwise.
    """
    support_chain_angular_domain = support_chain.get_dot_angle_values()
    ext_cad_1 = ch_j.extA if endpoint == EndPoints.A else ch_j.extB
    ext_cad_2 = (
        candidate_chain.extB if endpoint == EndPoints.A else candidate_chain.extA
    )
    interpolation_domain = compute_interpolation_domain(
        endpoint, ext_cad_1, ext_cad_2, support_chain.nr
    )
    intersection = np.intersect1d(interpolation_domain, support_chain_angular_domain)

    return True if len(intersection) == len(interpolation_domain) else False


def connectivity_goodness_condition(
    state: ChainSystemManager,
    ch_j: Chain,
    candidate_chain: Chain,
    ch_i: Chain,
    endpoint: int,
) -> Tuple[bool, float]:
    """
    Check if the chain candidate_chain can be connected to the chain ch_j.
    Implements Algorithm 4 of the paper.

    Args:
        state (SystemStatus): System status.
        ch_j (Chain): Chain j.
        candidate_chain (Chain): Candidate chain.
        ch_i (Chain): Support chain.
        endpoint (int): ch_j endpoint.

    Returns:
        Tuple[bool, float]: True if the chain candidate_chain can be connected to the chain ch_j, and the radial distance.
    """
    # Size criterion
    if ch_j.size + candidate_chain.size > ch_j.nr:
        return (False, -1)

    # Connect chains by correct endpoint
    check_pass = check_endpoints(ch_i, ch_j, candidate_chain, endpoint)
    if not check_pass:
        return (False, -1)

    # Radial check
    check_pass, distribution_distance = similarity_conditions(
        state,
        state.th_radial_tolerance,
        state.th_distribution_size,
        state.th_regular_derivative,
        state.derivative_from_center,
        ch_i,
        ch_j,
        candidate_chain,
        endpoint,
    )

    return (check_pass, distribution_distance)


def get_ids_chain_intersection(state: ChainSystemManager, chain_id: int) -> List[int]:
    """
    Get the IDs of chains that intersect with the given chain.

    Args:
        state (SystemStatus): System status.
        chain_id (int): ID of the chain.

    Returns:
        List[int]: List of IDs of intersecting chains.
    """
    ids_interseccion = list(np.where(state.M[chain_id] == 1)[0])
    ids_interseccion.remove(chain_id)

    return ids_interseccion


def distance_between_border(chain_1: Chain, chain_2: Chain, border_1: int) -> float:
    """
    Calculate the distance between the borders of two chains.

    Args:
        chain_1 (Chain): First chain.
        chain_2 (Chain): Second chain.
        border_1 (int): Border of the first chain.

    Returns:
        float: Distance between the borders.
    """
    node1 = chain_1.extA if border_1 == EndPoints.A else chain_2.extB
    node2 = chain_2.extB if border_1 == EndPoints.A else chain_2.extA

    d = euclidean_distance_between_nodes(node1, node2)

    return d


def get_inward_and_outward_visible_chains(
    chain_list: List[Chain], chain: Chain, endpoint: int
) -> Tuple[Optional[Chain], Optional[Chain], Node]:
    """
    Get the inward and outward visible chains from the given chain and endpoint.

    Args:
        chain_list (List[Chain]): List of chains.
        chain (Chain): Reference chain.
        endpoint (int): Endpoint of the reference chain.

    Returns:
        Tuple[Optional[Chain], Optional[Chain], Node]: Inward chain, outward chain, and node direction.
    """
    node_direction = chain.extA if endpoint == EndPoints.A else chain.extB
    inward_chain = None
    outward_chain = None

    dot_chain_index, dots_over_ray_direction = get_dots_in_radial_direction(
        node_direction, chain_list
    )
    if dot_chain_index < 0:
        return None, None, node_direction

    if dot_chain_index > 0:
        down_dot = dots_over_ray_direction[dot_chain_index - 1]
        inward_chain = get_chain_from_list_by_id(chain_list, down_dot.chain_id)

    if len(dots_over_ray_direction) - 1 > dot_chain_index:
        up_dot = dots_over_ray_direction[dot_chain_index + 1]
        outward_chain = get_chain_from_list_by_id(chain_list, up_dot.chain_id)

    return inward_chain, outward_chain, node_direction


def get_dots_in_radial_direction(
    node_direction: Node, chain_list: List[Chain]
) -> Tuple[int, List[Node]]:
    """
    Get the dots in the radial direction from the given node direction.

    Args:
        node_direction (Node): Node direction.
        chain_list (List[Chain]): List of chains.

    Returns:
        Tuple[int, List[Node]]: Index of the dot chain and list of nodes over the ray direction.
    """
    chains_in_radial_direction = get_chains_within_angle(
        node_direction.angle, chain_list
    )
    nodes_over_ray = get_closest_dots_to_angle_on_radial_direction_sorted_by_ascending_distance_to_center(
        chains_in_radial_direction, node_direction.angle
    )

    list_dot_chain_index = [
        idx
        for idx, node in enumerate(nodes_over_ray)
        if node.chain_id == node_direction.chain_id
    ]
    if len(list_dot_chain_index) > 0:
        dot_chain_index = list_dot_chain_index[0]
    else:
        nodes_over_ray = []
        dot_chain_index = -1

    return dot_chain_index, nodes_over_ray


def update_chain_after_connect(
    state: ChainSystemManager, ch_j: Chain, ch_k: Chain
) -> int:
    """
    Update the chain information after connecting two chains.

    Args:
        state (SystemStatus): System status.
        ch_j (Chain): Chain j.
        ch_k (Chain): Chain k.

    Returns:
        int: Status code (0 for success).
    """
    for chain in state.l_ch_s:
        if chain.A_outward is not None:
            if chain.A_outward.id == ch_k.id:
                chain.A_outward = ch_j
        if chain.A_inward is not None:
            if chain.A_inward.id == ch_k.id:
                chain.A_inward = ch_j

        if chain.B_outward is not None:
            if chain.B_outward.id == ch_k.id:
                chain.B_outward = ch_j

        if chain.B_inward is not None:
            if chain.B_inward.id == ch_k.id:
                chain.B_inward = ch_j

    return 0


def intersection_between_chains(chain1: Chain, chain2: Chain) -> bool:
    """
    Check if there is an intersection between two chains.

    Args:
        chain1 (Chain): First chain.
        chain2 (Chain): Second chain.

    Returns:
        bool: True if there is an intersection, False otherwise.
    """
    angle_intersection = [
        node.angle for node in chain1.l_nodes if chain2.get_node_by_angle(node.angle)
    ]

    return True if len(angle_intersection) > 0 else False


def compute_intersection_matrix(
    chains_list: List[Chain], nodes_list: List[Node]
) -> np.ndarray:
    """
    Compute intersection matrix. If chain_i intersection chain_j then img_height[i,j] == img_height[j,i] == 1 else 0

    Args:
        chains_list: chains list
        nodes_list: nodes list
        nr: total rays in disk

    Returns:
        img_height: Square matrix of lenght len(l_ch_s).
    """
    m = np.eye(len(chains_list))

    for angle in np.arange(0, 360, 360 / config.nr):
        chains_id_over_direction = np.unique(
            [node.chain_id for node in nodes_list if node.angle == angle]
        )
        x, y = np.meshgrid(chains_id_over_direction, chains_id_over_direction)
        m[x, y] = 1

    return m



from typing import Tuple, List, Dict, Any
import numpy as np
import logging

from ..geometry.curve import Curve
from ..geometry.chain import Chain, TypeChains
from ..processing.preprocessing import resize_image_using_pil_lib
from ..utils.file_utils import write_json
from ..geometry.geometry_utils import visualize_chains_over_image
from ..config import config

logger = logging.getLogger(__name__)


def save_results(
    img_in: np.ndarray,
    img_pre: np.ndarray,
    devernay_edges: np.ndarray,
    devernay_curves_f: List[Curve],
    devernay_curves_s: List[Chain],
    devernay_curves_c: List[Chain],
    devernay_curves_p: List[Chain],
) -> None:
    """Save detection results to disk."""
    # Convert chains to labelme format
    labelme_data = chain_to_labelme(img_in, chain_list=devernay_curves_p)
    json_path = config.output_dir / "labelme.json"
    write_json(labelme_data, json_path)
    logger.info(f"Saved labelme JSON to {json_path}")

    # Resize if necessary
    m, n, _ = img_in.shape
    m_n, n_n = img_pre.shape
    if m != m_n:
        img_in = resize_image_using_pil_lib(img_in, m_n, n_n)

    # Save visualization images
    visualizations = {
        "input.png": (img_in, {}),
        "preprocessing.png": (img_pre, {}),
        "edges.png": (img_pre, {"devernay": devernay_edges}),
        "filter.png": (img_pre, {"filter": devernay_curves_f}),
        "chains.png": (img_in, {"chain_list": devernay_curves_s}),
        "connect.png": (img_in, {"chain_list": devernay_curves_c}),
        "postprocessing.png": (img_in, {"chain_list": devernay_curves_p}),
        "output.png": (
            img_in,
            {
                "chain_list": [
                    chain
                    for chain in devernay_curves_p
                    if chain.is_closed()
                    and chain.type not in [TypeChains.center, TypeChains.border]
                ]
            },
        ),
    }

    for filename, (img, kwargs) in visualizations.items():
        output_path = config.output_dir / filename
        visualize_chains_over_image(img=img, filename=str(output_path), **kwargs)
        logger.debug(f"Saved visualization to {output_path}")


def chain_to_labelme(img_in: np.ndarray, chain_list: List[Chain]) -> Dict[str, Any]:
    """
    Convert chains to labelme format.
    The JSON is formatted to use the input image as the background and the chains as polygons.
    """
    init_height, init_width, _ = img_in.shape

    completed_chains = [
        chain
        for chain in chain_list
        if chain.is_closed()
        and chain.type not in [TypeChains.center, TypeChains.border]
    ]

    width_cte = init_width / config.output_width if config.output_width != None else 1
    height_cte = (
        init_height / config.output_height if config.output_height != None else 1
    )

    labelme_json = {
        "imagePath": str(config.input_image),
        "imageHeight": init_height,
        "imageWidth": init_width,
        "version": "5.5.0",
        "flags": {},
        "shapes": [],
        "imageData": None,
        "center": [
            (config.cy * height_cte) if config.cy is not None else None,
            (config.cx * width_cte) if config.cx is not None else None,
        ],
    }

    for idx, chain in enumerate(completed_chains):
        ring = {
            "label": str(idx + 1),
            "points": [
                [node.x * width_cte, node.y * height_cte] for node in chain.l_nodes
            ],
            "shape_type": "polygon",
            "flags": {},
        }
        labelme_json["shapes"].append(ring)

    return labelme_json



from enum import Enum
from typing import List, Optional, Tuple, Any

import matplotlib

matplotlib.use("Agg")  # non-interactive; Must be set before importing pyplot

import matplotlib.pyplot as plt
import numpy as np

from ..visualization.drawing import Drawing
from ..geometry.node import Node
from ..geometry.curve import Curve
from ..geometry.chain import Chain, TypeChains, EndPoints


def euclidean_distance(pix1: np.ndarray, pix2: np.ndarray) -> float:
    """
    Calculates the Euclidean distance between two pixels.

    Args:
        pix1 (np.ndarray): Coordinates of the first pixel.
        pix2 (np.ndarray): Coordinates of the second pixel.

    Returns:
        float: Euclidean distance between pix1 and pix2.
    """
    return np.sqrt((pix1[0] - pix2[0]) ** 2 + (pix1[1] - pix2[1]) ** 2)


def get_node_from_list_by_angle(dot_list: List[Node], angle: float) -> Optional[Node]:
    """
    Retrieves a node from a list by its angle.

    Args:
        dot_list (List[Node]): List of nodes to search.
        angle (float): Angle to match.

    Returns:
        Optional[Node]: The node with the specified angle, or None if not found.
    """
    try:
        dot = next(dot for dot in dot_list if (dot.angle == angle))
    except StopIteration:
        dot = None
    return dot


def get_chain_from_list_by_id(
    chain_list: List[Chain], chain_id: int
) -> Optional[Chain]:
    """
    Retrieves a chain from a list by its ID.

    Args:
        chain_list (List[Chain]): List of chains to search.
        chain_id (int): ID of the chain to find.

    Returns:
        Optional[Chain]: The chain with the specified ID, or None if not found.
    """
    try:
        chain_in_list = next(chain for chain in chain_list if (chain.id == chain_id))
    except StopIteration:
        chain_in_list = None

    return chain_in_list


def euclidean_distance_between_nodes(d1: Node, d2: Node) -> float:
    """
    Computes Euclidean distance between two nodes.

    Args:
        d1 (Node): First node.
        d2 (Node): Second node.

    Returns:
        float: Euclidean distance between the two nodes.
    """
    v1 = np.array([d1.x, d1.y], dtype=float)
    v2 = np.array([d2.x, d2.y], dtype=float)

    return euclidean_distance(v1, v2)


def copy_node(node: Node) -> Node:
    """
    Creates a copy of a node.

    Args:
        node (Node): Node to copy.

    Returns:
        Node: A new node with the same attributes as the original.
    """
    return Node(
        x=node.x,
        y=node.y,
        chain_id=node.chain_id,
        radial_distance=node.radial_distance,
        angle=node.angle,
    )


def copy_chain(chain: Chain) -> Chain:
    """
    Creates a copy of a chain.

    Args:
        chain (Chain): Chain to copy.

    Returns:
        Chain: A new chain with copied nodes.
    """
    aux_chain = Chain(
        chain.id,
        chain.nr,
        chain.center,
        chain.img_height,
        chain.img_width,
        type=chain.type,
    )
    aux_chain_node_list = [copy_node(node) for node in chain.l_nodes]
    aux_chain.add_nodes_list(aux_chain_node_list)
    return aux_chain


def angular_distance_between_endpoints(endpoint_j: Node, endpoint_k: Node) -> float:
    """
    Compute angular distance between endpoints.

    Args:
        endpoint_j (Node): Endpoint node j of chain ch_j.
        endpoint_k (Node): Endpoint node k of chain ch_k.

    Returns:
        float: Angular distance in degrees.
    """
    cte_degrees_in_a_circle = 360
    angular_distance = (
        endpoint_j.angle - endpoint_k.angle + cte_degrees_in_a_circle
    ) % cte_degrees_in_a_circle
    return angular_distance


def angular_distance_between_chains(
    ch_j: Chain, ch_k: Chain, endpoint_j_type: EndPoints
) -> float:
    """
    Compute angular distance between chains' endpoints.

    If endpoint_j == A then compute distance between chain.extA and ch_k.extB.
    Otherwise, compute distance between chain.extB and ch_k.extA.

    Args:
        ch_j (Chain): Chain j.
        ch_k (Chain): Chain k.
        endpoint_j_type (EndPoints): Endpoint type of chain j (A or B).

    Returns:
        float: Angular distance between endpoints in degrees.
    """
    endpoint_k = ch_k.extB if endpoint_j_type == EndPoints.A else ch_k.extA
    endpoint_j = ch_j.extA if endpoint_j_type == EndPoints.A else ch_j.extB

    angular_distance = (
        angular_distance_between_endpoints(endpoint_k, endpoint_j)
        if endpoint_j_type == EndPoints.B
        else angular_distance_between_endpoints(endpoint_j, endpoint_k)
    )

    return angular_distance


def minimum_euclidean_distance_between_vector_and_matrix(
    vector: np.ndarray, matrix: np.ndarray
) -> float:
    """
    Compute minimum Euclidean distance between a vector and each row in a matrix.

    Args:
        vector (np.ndarray): Vector.
        matrix (np.ndarray): Matrix.

    Returns:
        float: Minimum distance.
    """
    distances = np.sqrt(np.sum((matrix - vector) ** 2, axis=1))

    return np.min(distances)


def minimum_euclidean_distance_between_chains_endpoints(
    ch_j: Chain, ch_k: Chain
) -> float:
    """
    Compute minimum Euclidean distance between ch_j and ch_k endpoints.

    Args:
        ch_j (Chain): Chain j.
        ch_k (Chain): Chain k.

    Returns:
        float: Minimum Euclidean distance between endpoints.
    """
    nodes1, c1a, c1b = ch_j.to_array()
    nodes2, c2a, c2b = ch_k.to_array()
    c2a_min = minimum_euclidean_distance_between_vector_and_matrix(c2a, nodes1)
    c2b_min = minimum_euclidean_distance_between_vector_and_matrix(c2b, nodes1)
    c1a_min = minimum_euclidean_distance_between_vector_and_matrix(c1a, nodes2)
    c1b_min = minimum_euclidean_distance_between_vector_and_matrix(c1b, nodes2)
    return np.min([c2a_min, c2b_min, c1a_min, c1b_min])


def get_chains_within_angle(angle: float, chain_list: List[Chain]) -> List[Chain]:
    """
    Get chains that cover a specific angle.

    Args:
        angle (float): Angle in degrees.
        chain_list (List[Chain]): List of chains.

    Returns:
        List[Chain]: List of chains that cover the angle.
    """
    chains_list = []
    for chain in chain_list:
        A = chain.extA.angle
        B = chain.extB.angle
        if (A <= B and A <= angle <= B) or (A > B and (A <= angle or angle <= B)):
            chains_list.append(chain)
    return chains_list


def get_closest_chain_border_to_angle(chain: Chain, angle: float) -> Node:
    """
    Get the closest endpoint of a chain to a given angle.

    Args:
        chain (Chain): Chain to search.
        angle (float): Angle in degrees.

    Returns:
        Node: Closest endpoint node.
    """
    B = chain.extB.angle
    A = chain.extA.angle
    if B < A:
        dist_to_b = 360 - angle + B if angle > B else B - angle
        dist_to_a = angle - A if angle > B else 360 - A + angle
    else:
        dist_to_a = A - angle
        dist_to_b = angle - B
    # assert dist_to_a > 0 and dist_to_b > 0
    dot = chain.extB if dist_to_b < dist_to_a else chain.extA
    return dot


def get_closest_dots_to_angle_on_radial_direction_sorted_by_ascending_distance_to_center(
    chains_list: List[Chain], angle: float
) -> List[Node]:
    """
    Get nodes of all chains that are over the ray defined by angle and sort them by ascending distance to center.

    Args:
        chains_list (List[Chain]): List of chains.
        angle (float): Ray angle direction.

    Returns:
        List[Node]: Nodes sorted by ascending distance to center over ray direction angle.
    """
    node_list_over_ray = []
    for chain in chains_list:
        try:
            node = next(node for node in chain.l_nodes if node.angle == angle)
        except StopIteration:
            node = get_closest_chain_border_to_angle(chain, angle)

        if node not in node_list_over_ray:
            node_list_over_ray.append(node)

    if node_list_over_ray:
        node_list_over_ray.sort(key=lambda x: x.radial_distance)

    return node_list_over_ray


def get_nodes_from_chain_list(chain_list: List[Chain]) -> List[Node]:
    """
    Retrieves all nodes from a list of chains.

    Args:
        chain_list (List[Chain]): List of chains.

    Returns:
        List[Node]: List of all nodes in the chains.
    """
    inner_nodes = []
    for chain in chain_list:
        inner_nodes.extend(chain.l_nodes)
    return inner_nodes


def get_nodes_angles_from_list_nodes(node_list: List[Node]) -> List[float]:
    """
    Retrieves angles from a list of nodes.

    Args:
        node_list (List[Node]): List of nodes.

    Returns:
        List[float]: List of node angles.
    """
    return [node.angle for node in node_list]


def visualize_chains_over_image(
    chain_list: List[Chain] = [],
    img: Optional[np.ndarray] = None,
    filename: Optional[str] = None,
    devernay: Optional[np.ndarray] = None,
    filter: Optional[np.ndarray] = None,
) -> None:
    """
    Visualizes chains over an image.

    Args:
        chain_list (List[Chain]): List of chains to visualize.
        img (Optional[np.ndarray]): Image to display chains over.
        filename (Optional[str]): Filename to save the image.
        devernay (Optional[np.ndarray]): Devernay curves.
        filter (Optional[np.ndarray]): Filtered curves.

    Returns:
        None
    """
    if devernay is not None:
        img = Drawing.write_curves_to_image(devernay, img)
    elif filter is not None:
        img = write_filter_curves_to_image(filter, img)

    figsize = (10, 10)
    plt.figure(figsize=figsize)
    plt.imshow(img, cmap="gray")

    for chain in chain_list:
        x, y = chain.get_nodes_coordinates()
        if chain.type == TypeChains.normal:
            if chain.is_closed():
                x = x.tolist() + [x[0]]
                y = y.tolist() + [y[0]]
                plt.plot(x, y, "b", linewidth=1)
            else:
                plt.plot(x, y, "r", linewidth=1)
        elif chain.type == TypeChains.border:
            plt.plot(x, y, "k", linewidth=1)
        else:
            plt.scatter(int(x[0]), int(y[0]), c="k")

    plt.tight_layout()
    plt.axis("off")

    if filename:
        plt.savefig(filename)
    plt.close()


def visualize_selected_ch_and_chains_over_image_(
    selected_ch: List[Chain] = [],
    chain_list: List[Chain] = [],
    img: Optional[np.ndarray] = None,
    filename: Optional[str] = None,
    devernay: Optional[np.ndarray] = None,
    filter: Optional[np.ndarray] = None,
) -> None:
    """
    Visualizes selected chains and chain list over an image.

    Args:
        selected_ch (List[Chain]): List of selected chains to highlight.
        chain_list (List[Chain]): List of chains to visualize.
        img (Optional[np.ndarray]): Image to display chains over.
        filename (Optional[str]): Filename to save the image.
        devernay (Optional[np.ndarray]): Devernay curves.
        filter (Optional[np.ndarray]): Filtered curves.

    Returns:
        None
    """
    if devernay is not None:
        img = Drawing.write_curves_to_image(devernay, img)
    elif filter is not None:
        img = write_filter_curves_to_image(filter, img)

    figsize = (10, 10)
    plt.figure(figsize=figsize)
    plt.imshow(img, cmap="gray")
    for chain in chain_list:
        x, y = chain.get_nodes_coordinates()
        plt.plot(x, y, "w", linewidth=3)

    # draw selected chains
    for ch in selected_ch:
        x, y = ch.get_nodes_coordinates()
        plt.plot(x, y, linewidth=3)
        plt.annotate(str(ch.label_id), (x[0], y[0]), c="b")

    plt.tight_layout()
    plt.axis("off")
    if filename:
        plt.savefig(filename)
    plt.close()


def write_filter_curves_to_image(curves: List[Curve], img: np.ndarray) -> np.ndarray:
    """
    Draws filtered curves onto an image.

    Args:
        curves (List[Curve]): List of filtered curves.
        img (np.ndarray): Image to draw on.

    Returns:
        np.ndarray: Image with curves drawn.
    """
    img = np.full_like(img, 255)

    for c in curves:
        img = c.draw(img, thickness=3)

    return img



import matplotlib.pyplot as plt

from ..geometry.chain import Chain, EndPoints, ClockDirection
from ..geometry.node import Node
from ..analysis.interpolation_nodes import domain_interpolation
from ..geometry.geometry_utils import euclidean_distance_between_nodes

from typing import List


class ChainNeighbourhood:
    """
    Class to compute and store the total nodes of a chain and the candidate chains to connect to it.

    It generates the virtual nodes to compute the similarity condition.
    """

    def __init__(
        self,
        src_chain: Chain,
        dst_chain: Chain,
        support_chain: Chain,
        endpoint: EndPoints,
        n_nodes: int = 20,
    ):
        """
        Initializes a ChainNeighbourhood instance.

        Args:
            src_chain (Chain): Source chain to be connected.
            dst_chain (Chain): Destination chain to be connected.
            support_chain (Chain): Support chain to be used to generate the virtual nodes.
            endpoint (EndPoints): Endpoint of the source chain.
            n_nodes (int): Number of nodes to be considered in the total nodes.
        """
        self.virtual_nodes = self.generate_virtual_nodes_between_two_chains(
            src_chain, dst_chain, support_chain, endpoint
        )
        self.src_endpoint = (
            src_chain.extA if endpoint == EndPoints.A else src_chain.extB
        )
        self.dst_endpoint = (
            dst_chain.extA if endpoint == EndPoints.B else dst_chain.extB
        )
        self.endpoint_and_virtual_nodes = (
            [self.src_endpoint] + self.virtual_nodes + [self.dst_endpoint]
        )
        self.radial_distance_endpoints_to_support = (
            self.radial_distance_between_nodes_belonging_to_same_ray(
                [self.src_endpoint, self.dst_endpoint], support_chain
            )
        )
        self.radial_distance_virtual_nodes_to_support = (
            self.radial_distance_between_nodes_belonging_to_same_ray(
                self.virtual_nodes, support_chain
            )
        )

        self.src_chain_nodes = (
            src_chain.sort_dots(direction=ClockDirection.anti_clockwise)[:n_nodes]
            if endpoint == EndPoints.A
            else src_chain.sort_dots(direction=ClockDirection.clockwise)[:n_nodes]
        )
        self.dst_chain_nodes = (
            dst_chain.sort_dots(direction=ClockDirection.clockwise)[:n_nodes]
            if endpoint == EndPoints.A
            else dst_chain.sort_dots(direction=ClockDirection.anti_clockwise)[:n_nodes]
        )
        self.set_i = self.radial_distance_between_nodes_belonging_to_same_ray(
            self.src_chain_nodes, support_chain
        )
        self.set_k = self.radial_distance_between_nodes_belonging_to_same_ray(
            self.dst_chain_nodes, support_chain
        )

        if endpoint == EndPoints.B:
            self.neighbourhood_nodes = (
                self.src_chain_nodes[::-1] + self.virtual_nodes + self.dst_chain_nodes
            )
        else:
            self.neighbourhood_nodes = (
                self.dst_chain_nodes[::-1]
                + self.virtual_nodes[::-1]
                + self.src_chain_nodes
            )

    def draw_neighbourhood(self, name: str) -> None:
        """
        Draws the neighbourhood radial distances.

        Args:
            name (str): Filename to save the plot.
        """
        r = [node.radial_distance for node in self.neighbourhood_nodes]
        plt.figure()
        plt.plot(r, "b")
        plt.savefig(name)
        plt.close()

    def generate_virtual_nodes_between_two_chains(
        self,
        src_chain: Chain,
        dst_chain: Chain,
        support_chain: Chain,
        endpoint: EndPoints,
    ) -> List[Node]:
        """
        Generates virtual nodes between two chains using a support chain.

        Args:
            src_chain (Chain): Source chain to be connected.
            dst_chain (Chain): Destination chain to be connected.
            support_chain (Chain): Support chain used to generate the virtual nodes.
            endpoint (EndPoints): Endpoint of the source chain.

        Returns:
            List[Node]: List of virtual nodes.
        """
        virtual_nodes = []
        cad1_endpoint = src_chain.extA if endpoint == EndPoints.A else src_chain.extB
        cad2_endpoint = dst_chain.extB if endpoint == EndPoints.A else dst_chain.extA

        domain_interpolation(
            support_chain,
            cad1_endpoint,
            cad2_endpoint,
            endpoint,
            src_chain,
            virtual_nodes,
        )

        return virtual_nodes

    def radial_distance_between_nodes_belonging_to_same_ray(
        self, node_list: List[Node], support_chain: Chain
    ) -> List[float]:
        """
        Computes radial distances between nodes belonging to the same ray and a support chain.

        Args:
            node_list (List[Node]): List of nodes.
            support_chain (Chain): Support chain.

        Returns:
            List[float]: List of radial distances between nodes and support chain nodes.
        """
        radial_distances = []

        for node in node_list:
            support_node = support_chain.get_node_by_angle(node.angle)
            if support_node is None:
                break
            radial_distances.append(
                euclidean_distance_between_nodes(support_node, node)
            )

        return radial_distances



from typing import List, Tuple
from itertools import cycle


class Color:
    """Class representing various colors in BGR format."""

    yellow = (0, 255, 255)
    red = (0, 0, 255)
    blue = (255, 0, 0)
    dark_yellow = (0, 204, 204)
    cyan = (255, 255, 0)
    orange = (0, 165, 255)
    purple = (255, 0, 255)
    maroon = (34, 34, 178)
    green = (0, 255, 0)
    white = (255, 255, 255)
    black = (0, 0, 0)
    gray_white = 255
    gray_black = 0

    def __init__(self):
        """
        Initializes a Color instance with a cyclic iterator of colors.
        """
        self.list: List[Tuple[int, int, int]] = [
            self.yellow,
            self.red,
            self.blue,
            self.dark_yellow,
            self.cyan,
            self.orange,
            self.purple,
            self.maroon,
        ]
        self.color_cycle = cycle(self.list)

    def get_next_color(self) -> Tuple[int, int, int]:
        """
        Returns the next color in the cyclic iterator.

        Returns:
            Tuple[int, int, int]: The next color in BGR format.
        """
        return next(self.color_cycle)



import cv2
from PIL import Image
import numpy as np
from typing import Tuple
import logging

from ..config import config

logger = logging.getLogger(__name__)

WHITE = 255
NONE = 0


def get_image_shape(img_in: np.ndarray) -> Tuple[int, int]:
    """
    Get image shape.

    Args:
        img_in (np.ndarray): Input image.

    Returns:
        Tuple[int, int]: Image height, image width.
    """
    if img_in.ndim > 2:
        height, width, _ = img_in.shape
    else:
        height, width = img_in.shape

    return height, width


def resize(img_in: cv2.typing.MatLike) -> np.ndarray:
    """
    Resize image and keep the center of the image in the same position. Implements Algorithm 2 in the supplementary material.

    Args:
        img_in (cv2.typing.MatLike): Image to resize.

    Returns:
        np.ndarray: Resized image, resized y's center coordinate, resized x's center coordinate.
    """
    logger.debug("Resizing image")

    current_height, current_width = get_image_shape(img_in)

    logger.debug(f"Current image shape: {current_height}x{current_width}")

    # Calculate missing dimension if needed
    if config.output_height is not None and config.output_width is None:
        aspect_ratio = current_width / current_height
        width_output = int(config.output_height * aspect_ratio)
        config.update(output_width=width_output)
    elif config.output_width is not None and config.output_height is None:
        aspect_ratio = current_height / current_width
        height_output = int(config.output_width * aspect_ratio)
        config.update(output_height=height_output)

    logger.debug(f"Resizing image to {config.output_height}x{config.output_width}")

    img_resized = resize_image_using_pil_lib(
        img_in, config.output_width, config.output_height
    )
    convert_center_coordinate_to_output_coordinate(current_height, current_width)

    return img_resized


def resize_image_using_pil_lib(
    img_in: cv2.typing.MatLike, target_width, target_height
) -> np.ndarray:
    """
    Resize image using PIL library.

    Args:
        img_in (cv2.typing.MatLike): Input image.

    Returns:
        np.ndarray: Matrix with the resized image.
    """
    image_pil = Image.fromarray(img_in)
    image_pil = image_pil.resize(
        (target_width, target_height), Image.Resampling.LANCZOS
    )
    # Convert PIL image to numpy array
    image_pil_resized = np.array(image_pil)

    return image_pil_resized


def convert_center_coordinate_to_output_coordinate(
    input_height: int, input_width: int
) -> None:
    """
    Convert center coordinate from input image to output image.

    Args:
        input_height (int): Input image height.
        input_width (int): Input image width.

    Returns:
        None
    """
    hscale = config.output_height / input_height
    wscale = config.output_width / input_width

    cy_output = config.cy * hscale
    cx_output = config.cx * wscale

    config.update(cy=cy_output, cx=cx_output)


def change_background_intensity_to_mean(
    img_in: np.ndarray,
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Change background intensity to mean intensity.

    Args:
        img_in (np.ndarray): Input gray scale image. Background is white (255).

    Returns:
        Tuple[np.ndarray, np.ndarray]: Image with changed background intensity, background mask.
    """
    im_eq = img_in.copy()
    mask = np.where(img_in == 255, 1, 0)
    im_eq = change_background_to_value(im_eq, mask, np.mean(img_in[mask == 0]))

    return im_eq, mask


def equalize_image_using_clahe(img_eq: np.ndarray) -> np.ndarray:
    """
    Equalize image using CLAHE algorithm.

    Args:
        img_eq (np.ndarray): Image.

    Returns:
        np.ndarray: Equalized image.
    """
    clahe = cv2.createCLAHE(clipLimit=10)
    img_eq = clahe.apply(img_eq)

    return img_eq


def equalize(im_g: np.ndarray) -> np.ndarray:
    """
    Equalize image using CLAHE algorithm. Implements Algorithm 3 in the supplementary material.

    Args:
        im_gray (np.ndarray): Gray scale image.

    Returns:
        np.ndarray: Equalized image.
    """
    img_pre, mask = change_background_intensity_to_mean(im_g)
    img_pre = equalize_image_using_clahe(img_pre)
    img_pre = change_background_to_value(img_pre, mask, WHITE)

    return img_pre


def change_background_to_value(
    img_in: np.ndarray, mask: np.ndarray, value: int = 255
) -> np.ndarray:
    """
    Change background intensity to white.

    Args:
        img_in (np.ndarray): Input image.
        mask (np.ndarray): Background mask.
        value (int): Value to change the background to.

    Returns:
        np.ndarray: Image with changed background intensity.
    """
    img_in[mask > 0] = value

    return img_in


def rgb2gray(img_r: np.ndarray) -> np.ndarray:
    """
    Convert RGB image to grayscale.

    Args:
        img_r (np.ndarray): RGB image.

    Returns:
        np.ndarray: Grayscale image.
    """
    return cv2.cvtColor(img_r, cv2.COLOR_BGR2GRAY)


def preprocessing(img_in: cv2.typing.MatLike) -> np.ndarray:
    """
    Image preprocessing steps. Following actions are made:
    - Image resize
    - Image is converted to gray scale
    - Gray scale image is equalized
    Implements Algorithm 1 in the supplementary material.

    Args:
        img_in (cv2.typing.MatLike): Segmented image.

    Returns:
        np.ndarray: Equalized image, pith y's coordinate after resize, pith x's coordinate after resize.
    """
    if config.output_height is None and config.output_width is None:
        img_resized = img_in
    else:
        img_resized = resize(img_in)

    im_gray = rgb2gray(img_resized)
    img_pre = equalize(im_gray)

    return img_pre



from typing import List, Tuple
import numpy as np
import cv2
from shapely.geometry import LineString
from ..visualization.color import Color


class Curve:
    def __init__(self, pixels_list: List[Tuple[int, int]], name: int):
        """
        Initializes a Curve object.

        Args:
            pixels_list (List[Tuple[int, int]]): List of pixel coordinates (y, x).
            name (int): Identifier of the curve.
        """
        self.id = name
        self.geometry = LineString(np.array(pixels_list)[:, [1, 0]].tolist())

    def draw(self, img: np.ndarray, thickness: int = 1) -> np.ndarray:
        """
        Draws the curve on an image.

        Args:
            img (np.ndarray): Image on which to draw.
            thickness (int): Thickness of the curve lines.

        Returns:
            np.ndarray: Image with the curve drawn.
        """
        y, x = self.geometry.xy
        y = np.array(y).astype(int)
        x = np.array(x).astype(int)
        pts = np.vstack((x, y)).T
        isClosed = False
        img = cv2.polylines(img, [pts], isClosed, Color.black, thickness)

        return img
